# Copyright (c) 2014 Ben Swartzlander.  All rights reserved.
# Copyright (c) 2014 Navneet Singh.  All rights reserved.
# Copyright (c) 2014 Clinton Knight.  All rights reserved.
# Copyright (c) 2014 Alex Meade.  All rights reserved.
# Copyright (c) 2014 Bob Callaway.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""
Tests for NetApp API layer
"""
import ddt
import mock
from six.moves import urllib

from manila import exception
from manila.share.drivers.netapp.dataontap.client import api
from manila import test
from manila.tests.share.drivers.netapp.dataontap.client import fakes as fake


class NetAppApiElementTransTests(test.TestCase):
    """Test case for NetApp API element translations."""

    def test_get_set_system_version(self):
        napi = api.NaServer('localhost')

        # Testing calls before version is set
        version = napi.get_system_version()
        self.assertIsNone(version)
        napi.set_system_version(fake.VERSION_TUPLE)
        version = napi.get_system_version()
        self.assertEqual(fake.VERSION_TUPLE, version)

    def test_translate_struct_dict_unique_key(self):
        """Tests if dict gets properly converted to NaElements."""
        root = api.NaElement('root')
        child = {'e1': 'v1', 'e2': 'v2', 'e3': 'v3'}

        root.translate_struct(child)

        self.assertEqual(3, len(root.get_children()))
        for key, value in child.items():
            self.assertEqual(value, root.get_child_content(key))

    def test_translate_struct_dict_nonunique_key(self):
        """Tests if list/dict gets properly converted to NaElements."""
        root = api.NaElement('root')
        child = [{'e1': 'v1', 'e2': 'v2'}, {'e1': 'v3'}]

        root.translate_struct(child)

        children = root.get_children()
        self.assertEqual(3, len(children))
        for c in children:
            if c.get_name() == 'e1':
                self.assertIn(c.get_content(), ['v1', 'v3'])
            else:
                self.assertEqual('v2', c.get_content())

    def test_translate_struct_list(self):
        """Tests if list gets properly converted to NaElements."""
        root = api.NaElement('root')
        child = ['e1', 'e2']

        root.translate_struct(child)

        self.assertEqual(2, len(root.get_children()))
        self.assertIsNone(root.get_child_content('e1'))
        self.assertIsNone(root.get_child_content('e2'))

    def test_translate_struct_tuple(self):
        """Tests if tuple gets properly converted to NaElements."""
        root = api.NaElement('root')
        child = ('e1', 'e2')

        root.translate_struct(child)

        self.assertEqual(2, len(root.get_children()))
        self.assertIsNone(root.get_child_content('e1'))
        self.assertIsNone(root.get_child_content('e2'))

    def test_translate_invalid_struct(self):
        """Tests if invalid data structure raises exception."""
        root = api.NaElement('root')
        child = 'random child element'
        self.assertRaises(ValueError, root.translate_struct, child)

    def test_setter_builtin_types(self):
        """Tests str, int, float get converted to NaElement."""
        update = dict(e1='v1', e2='1', e3='2.0', e4='8')
        root = api.NaElement('root')

        for key, value in update.items():
            root[key] = value

        for key, value in update.items():
            self.assertEqual(value, root.get_child_content(key))

    def test_setter_na_element(self):
        """Tests na_element gets appended as child."""
        root = api.NaElement('root')
        root['e1'] = api.NaElement('nested')
        self.assertEqual(1, len(root.get_children()))
        e1 = root.get_child_by_name('e1')
        self.assertIsInstance(e1, api.NaElement)
        self.assertIsInstance(e1.get_child_by_name('nested'), api.NaElement)

    def test_setter_child_dict(self):
        """Tests dict is appended as child to root."""
        root = api.NaElement('root')
        root['d'] = {'e1': 'v1', 'e2': 'v2'}
        e1 = root.get_child_by_name('d')
        self.assertIsInstance(e1, api.NaElement)
        sub_ch = e1.get_children()
        self.assertEqual(2, len(sub_ch))
        for c in sub_ch:
            self.assertIn(c.get_name(), ['e1', 'e2'])
            if c.get_name() == 'e1':
                self.assertEqual('v1', c.get_content())
            else:
                self.assertEqual('v2', c.get_content())

    def test_setter_child_list_tuple(self):
        """Tests list/tuple are appended as child to root."""
        root = api.NaElement('root')

        root['l'] = ['l1', 'l2']
        root['t'] = ('t1', 't2')

        l = root.get_child_by_name('l')
        self.assertIsInstance(l, api.NaElement)
        t = root.get_child_by_name('t')
        self.assertIsInstance(t, api.NaElement)

        self.assertEqual(2, len(l.get_children()))
        for le in l.get_children():
            self.assertIn(le.get_name(), ['l1', 'l2'])

        self.assertEqual(2, len(t.get_children()))
        for te in t.get_children():
            self.assertIn(te.get_name(), ['t1', 't2'])

    def test_setter_no_value(self):
        """Tests key with None value."""
        root = api.NaElement('root')
        root['k'] = None
        self.assertIsNone(root.get_child_content('k'))

    def test_setter_invalid_value(self):
        """Tests invalid value raises exception."""
        self.assertRaises(TypeError,
                          api.NaElement('root').__setitem__,
                          'k',
                          api.NaServer('localhost'))

    def test_setter_invalid_key(self):
        """Tests invalid value raises exception."""
        self.assertRaises(KeyError,
                          api.NaElement('root').__setitem__,
                          None,
                          'value')


@ddt.ddt
class NetAppApiServerTests(test.TestCase):
    """Test case for NetApp API server methods"""
    def setUp(self):
        self.root = api.NaServer('127.0.0.1')
        super(NetAppApiServerTests, self).setUp()

    @ddt.data(None, fake.FAKE_XML_STR)
    def test_invoke_elem_value_error(self, na_element):
        """Tests whether invalid NaElement parameter causes error"""

        self.assertRaises(ValueError, self.root.invoke_elem, na_element)

    def test_invoke_elem_http_error(self):
        """Tests handling of HTTPError"""
        na_element = fake.FAKE_NA_ELEMENT
        self.mock_object(self.root, '_create_request', mock.Mock(
            return_value=('abc', fake.FAKE_NA_ELEMENT)))
        self.mock_object(api, 'LOG')
        self.root._opener = fake.FAKE_HTTP_OPENER
        self.mock_object(self.root, '_build_opener')
        self.mock_object(self.root._opener, 'open', mock.Mock(
            side_effect=urllib.error.HTTPError(url='', hdrs='',
                                               fp=None, code='401',
                                               msg='httperror')))

        self.assertRaises(api.NaApiError, self.root.invoke_elem,
                          na_element)

    def test_invoke_elem_urlerror(self):
        """Tests handling of URLError"""
        na_element = fake.FAKE_NA_ELEMENT
        self.mock_object(self.root, '_create_request', mock.Mock(
            return_value=('abc', fake.FAKE_NA_ELEMENT)))
        self.mock_object(api, 'LOG')
        self.root._opener = fake.FAKE_HTTP_OPENER
        self.mock_object(self.root, '_build_opener')
        self.mock_object(self.root._opener, 'open', mock.Mock(
            side_effect=urllib.error.URLError(reason='urlerror')))

        self.assertRaises(exception.StorageCommunicationException,
                          self.root.invoke_elem,
                          na_element)

    def test_invoke_elem_unknown_exception(self):
        """Tests handling of Unknown Exception"""
        na_element = fake.FAKE_NA_ELEMENT
        self.mock_object(self.root, '_create_request', mock.Mock(
            return_value=('abc', fake.FAKE_NA_ELEMENT)))
        self.mock_object(api, 'LOG')
        self.root._opener = fake.FAKE_HTTP_OPENER
        self.mock_object(self.root, '_build_opener')
        self.mock_object(self.root._opener, 'open', mock.Mock(
            side_effect=Exception))

        exception = self.assertRaises(api.NaApiError, self.root.invoke_elem,
                                      na_element)
        self.assertEqual('unknown', exception.code)

    @ddt.data({'trace_enabled': False,
               'trace_pattern': '(.*)', 'log': False},
              {'trace_enabled': True,
               'trace_pattern': '(?!(volume)).*', 'log': False},
              {'trace_enabled': True,
               'trace_pattern': '(.*)', 'log': True},
              {'trace_enabled': True,
               'trace_pattern': '^volume-(info|get-iter)$', 'log': True})
    @ddt.unpack
    def test_invoke_elem_valid(self, trace_enabled, trace_pattern, log):
        """Tests the method invoke_elem with valid parameters"""
        na_element = fake.FAKE_NA_ELEMENT
        self.root._trace = trace_enabled
        self.root._api_trace_pattern = trace_pattern
        self.mock_object(self.root, '_create_request', mock.Mock(
            return_value=('abc', fake.FAKE_NA_ELEMENT)))
        self.mock_object(api, 'LOG')
        self.root._opener = fake.FAKE_HTTP_OPENER
        self.mock_object(self.root, '_build_opener')
        self.mock_object(self.root, '_get_result', mock.Mock(
            return_value=fake.FAKE_NA_ELEMENT))
        opener_mock = self.mock_object(
            self.root._opener, 'open', mock.Mock())
        opener_mock.read.side_effect = ['resp1', 'resp2']

        self.root.invoke_elem(na_element)

        expected_log_count = 2 if log else 0
        self.assertEqual(expected_log_count, api.LOG.debug.call_count)



# Copyright (c) 2016 Hitachi Data Systems, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy
import ddt
import mock

from manila.common import constants
from manila import context
from manila import db
from manila import exception
from manila.share import snapshot_access
from manila import test
from manila.tests import db_utils
from manila import utils


@ddt.ddt
class SnapshotAccessTestCase(test.TestCase):
    def setUp(self):
        super(SnapshotAccessTestCase, self).setUp()
        self.driver = self.mock_class("manila.share.driver.ShareDriver",
                                      mock.Mock())
        self.snapshot_access = snapshot_access.ShareSnapshotInstanceAccess(
            db, self.driver)
        self.context = context.get_admin_context()
        share = db_utils.create_share()
        self.snapshot = db_utils.create_snapshot(share_id=share['id'])
        self.snapshot_instance = db_utils.create_snapshot_instance(
            snapshot_id=self.snapshot['id'],
            share_instance_id=self.snapshot['share']['instance']['id'])

    @ddt.data(constants.ACCESS_STATE_QUEUED_TO_APPLY,
              constants.ACCESS_STATE_QUEUED_TO_DENY)
    def test_update_access_rules(self, state):

        rules = []
        for i in range(2):
            rules.append({
                'id': 'id-%s' % i,
                'state': state,
                'access_id': 'rule_id%s' % i
            })
        all_rules = copy.deepcopy(rules)
        all_rules.append({
            'id': 'id-3',
            'state': constants.ACCESS_STATE_ERROR,
            'access_id': 'rule_id3'
        })

        snapshot_instance_get = self.mock_object(
            db, 'share_snapshot_instance_get',
            mock.Mock(return_value=self.snapshot_instance))

        snap_get_all_for_snap_instance = self.mock_object(
            db, 'share_snapshot_access_get_all_for_snapshot_instance',
            mock.Mock(return_value=all_rules))

        self.mock_object(db, 'share_snapshot_instance_access_update')
        self.mock_object(self.driver, 'snapshot_update_access')
        self.mock_object(self.snapshot_access, '_check_needs_refresh',
                         mock.Mock(return_value=False))
        self.mock_object(db, 'share_snapshot_instance_access_delete')

        self.snapshot_access.update_access_rules(self.context,
                                                 self.snapshot_instance['id'])

        snapshot_instance_get.assert_called_once_with(
            utils.IsAMatcher(context.RequestContext),
            self.snapshot_instance['id'], with_share_data=True)
        snap_get_all_for_snap_instance.assert_called_once_with(
            utils.IsAMatcher(context.RequestContext),
            self.snapshot_instance['id'])
        if state == constants.ACCESS_STATE_QUEUED_TO_APPLY:
            self.driver.snapshot_update_access.assert_called_once_with(
                utils.IsAMatcher(context.RequestContext),
                self.snapshot_instance, rules, add_rules=rules,
                delete_rules=[], share_server=None)
        else:
            self.driver.snapshot_update_access.assert_called_once_with(
                utils.IsAMatcher(context.RequestContext),
                self.snapshot_instance, [], add_rules=[],
                delete_rules=rules, share_server=None)

    def test_update_access_rules_delete_all_rules(self):

        rules = []
        for i in range(2):
            rules.append({
                'id': 'id-%s' % i,
                'state': constants.ACCESS_STATE_QUEUED_TO_DENY,
                'access_id': 'rule_id%s' % i
            })

        snapshot_instance_get = self.mock_object(
            db, 'share_snapshot_instance_get',
            mock.Mock(return_value=self.snapshot_instance))

        snap_get_all_for_snap_instance = self.mock_object(
            db, 'share_snapshot_access_get_all_for_snapshot_instance',
            mock.Mock(side_effect=[rules, []]))

        self.mock_object(db, 'share_snapshot_instance_access_update')
        self.mock_object(self.driver, 'snapshot_update_access')
        self.mock_object(db, 'share_snapshot_instance_access_delete')

        self.snapshot_access.update_access_rules(self.context,
                                                 self.snapshot_instance['id'],
                                                 delete_all_rules=True)

        snapshot_instance_get.assert_called_once_with(
            utils.IsAMatcher(context.RequestContext),
            self.snapshot_instance['id'], with_share_data=True)
        snap_get_all_for_snap_instance.assert_called_with(
            utils.IsAMatcher(context.RequestContext),
            self.snapshot_instance['id'])
        self.driver.snapshot_update_access.assert_called_with(
            utils.IsAMatcher(context.RequestContext), self.snapshot_instance,
            [], add_rules=[], delete_rules=rules, share_server=None)

    def test_update_access_rules_exception(self):

        rules = []
        for i in range(2):
            rules.append({
                'id': 'id-%s' % i,
                'state': constants.ACCESS_STATE_APPLYING,
                'access_id': 'rule_id%s' % i
            })

        snapshot_instance_get = self.mock_object(
            db, 'share_snapshot_instance_get',
            mock.Mock(return_value=self.snapshot_instance))

        snap_get_all_for_snap_instance = self.mock_object(
            db, 'share_snapshot_access_get_all_for_snapshot_instance',
            mock.Mock(return_value=rules))

        self.mock_object(db, 'share_snapshot_instance_access_update')
        self.mock_object(self.driver, 'snapshot_update_access',
                         mock.Mock(side_effect=exception.NotFound))

        self.assertRaises(exception.NotFound,
                          self.snapshot_access.update_access_rules,
                          self.context, self.snapshot_instance['id'])

        snapshot_instance_get.assert_called_once_with(
            utils.IsAMatcher(context.RequestContext),
            self.snapshot_instance['id'], with_share_data=True)
        snap_get_all_for_snap_instance.assert_called_once_with(
            utils.IsAMatcher(context.RequestContext),
            self.snapshot_instance['id'])

        self.driver.snapshot_update_access.assert_called_once_with(
            utils.IsAMatcher(context.RequestContext), self.snapshot_instance,
            rules, add_rules=rules, delete_rules=[], share_server=None)



# Copyright 2016 Nexenta Systems, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from mock import patch
from oslo_serialization import jsonutils
import requests

from manila import exception
from manila.share.drivers.nexenta.ns5 import jsonrpc
from manila import test

PATH_TO_RPC = 'manila.share.drivers.nexenta.ns5.jsonrpc.NexentaJSONProxy'


class TestNexentaJSONProxy(test.TestCase):

    def __init__(self, method):
        super(TestNexentaJSONProxy, self).__init__(method)

    def setUp(self):
        super(TestNexentaJSONProxy, self).setUp()
        self.nef_get = jsonrpc.NexentaJSONProxy(
            'http', '1.1.1.1', '8080', 'user', 'pass', 'get')
        self.nef_post = jsonrpc.NexentaJSONProxy(
            'https', '1.1.1.1', '8080', 'user', 'pass', 'post')

    @patch('requests.Response.close')
    @patch('requests.Session.get')
    def test_call_get_data(self, get, close):
        data = {'key': 'value'}
        get.return_value = requests.Response()
        get.return_value.__setstate__(
            {'status_code': 200, '_content': jsonutils.dumps(data)})

        self.assertEqual({'key': 'value'}, self.nef_get('url'))

    @patch('requests.Response.close')
    @patch('requests.Session.get')
    def test_call_get_created(self, get, close):
        get.return_value = requests.Response()
        get.return_value.__setstate__({
            'status_code': 201, '_content': ''})

        self.assertIsNone(self.nef_get('url'))

    @patch('requests.Response.close')
    @patch('requests.Session.post')
    def test_call_post_success(self, post, close):
        data = {'key': 'value'}
        post.return_value = requests.Response()
        post.return_value.__setstate__({
            'status_code': 200, '_content': ''})
        self.assertIsNone(self.nef_post('url', data))

    @patch('time.sleep')
    @patch('requests.Response.close')
    @patch('requests.Session.get')
    @patch('requests.Session.post')
    def test_call_post_202(self, post, get, close, sleep):
        data = {'key': 'value'}
        data2 = {'links': [{'href': 'redirect_url'}]}

        get.return_value = requests.Response()
        post.return_value = requests.Response()
        post.return_value.__setstate__({
            'status_code': 202, '_content': jsonutils.dumps(data2)})
        get.return_value.__setstate__({
            'status_code': 200, '_content': jsonutils.dumps(data)})

        self.assertEqual({'key': 'value'}, self.nef_post('url'))

    @patch('requests.Response.close')
    @patch('requests.Session.get')
    def test_call_get_not_exist(self, get, close):
        get.return_value = requests.Response()
        get.return_value.__setstate__({
            'status_code': 400,
            '_content': jsonutils.dumps({'code': 'ENOENT'})})

        self.assertRaises(
            exception.NexentaException, lambda: self.nef_get('url'))

    @patch('requests.Response.close')
    @patch('requests.Session.get')
    def test_call_get_unauthorized(self, get, close):
        get.return_value = requests.Response()
        get.return_value.__setstate__({
            'status_code': 401,
            '_content': jsonutils.dumps({'code': 'unauthorized'})})

        self.assertRaises(
            exception.NexentaException, lambda: self.nef_get('url'))

    @patch('%s.https_auth' % PATH_TO_RPC)
    @patch('requests.Response.close')
    @patch('requests.Session.post')
    def test_call_post_bad_token(self, post, close, auth):
        post.return_value = requests.Response()
        auth.return_value = {'token': 'tok'}
        post.return_value.__setstate__({
            'status_code': 401,
            '_content': jsonutils.dumps({'code': 'unauthorized'})})

        self.assertRaises(
            exception.NexentaException, lambda: self.nef_post('url'))

    @patch('requests.Response.close')
    @patch('requests.Session.post')
    def test_auth(self, post, close):
        httpsdata = {'token': 'tok'}
        post.return_value = requests.Response()
        post.return_value.__setstate__({
            'status_code': 200, '_content': jsonutils.dumps(httpsdata)})
        nef_get = jsonrpc.NexentaJSONProxy(
            'http', '1.1.1.1', '8080', 'user', 'pass', method='get')
        https_auth = nef_get.https_auth()
        self.assertEqual('tok', https_auth)



#!/usr/bin/env python

# Copyright (c) 2012 Rackspace Hosting
# Copyright (c) 2013 NetApp
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Configuration support for all drivers.

This module allows support for setting configurations either from default
or from a particular CONF group, to be able to set multiple configurations
for a given set of values.

For instance, two generic configurations can be set by naming them in groups as

 [generic1]
 share_backend_name=generic-backend-1
 ...

 [generic2]
 share_backend_name=generic-backend-2
 ...

And the configuration group name will be passed in so that all calls to
configuration.volume_group within that instance will be mapped to the proper
named group.

This class also ensures the implementation's configuration is grafted into the
option group. This is due to the way cfg works. All cfg options must be defined
and registered in the group in which they are used.
"""

from oslo_config import cfg

CONF = cfg.CONF


class Configuration(object):

    def __init__(self, share_opts, config_group=None):
        """Graft config values into config group.

        This takes care of grafting the implementation's config values
        into the config group.
        """
        self.config_group = config_group

        # set the local conf so that __call__'s know what to use
        if self.config_group:
            self._ensure_config_values(share_opts)
            self.local_conf = CONF._get(self.config_group)
        else:
            self.local_conf = CONF

    def _ensure_config_values(self, share_opts):
        CONF.register_opts(share_opts,
                           group=self.config_group)

    def append_config_values(self, share_opts):
        self._ensure_config_values(share_opts)

    def safe_get(self, value):
        try:
            return self.__getattr__(value)
        except cfg.NoSuchOptError:
            return None

    def __getattr__(self, value):
        return getattr(self.local_conf, value)



# Copyright 2015, Hitachi Data Systems.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import sys

from manila.cmd import data as manila_data
from manila import test
from manila import version

CONF = manila_data.CONF


class ManilaCmdDataTestCase(test.TestCase):

    def test_main(self):
        sys.argv = ['manila-data']
        self.mock_object(manila_data.log, 'setup')
        self.mock_object(manila_data.log, 'register_options')
        self.mock_object(manila_data.utils, 'monkey_patch')
        self.mock_object(manila_data.service.Service, 'create')
        self.mock_object(manila_data.service, 'serve')
        self.mock_object(manila_data.service, 'wait')

        manila_data.main()

        self.assertEqual('manila', CONF.project)
        self.assertEqual(version.version_string(), CONF.version)
        manila_data.log.setup.assert_called_once_with(CONF, "manila")
        manila_data.log.register_options.assert_called_once_with(CONF)
        manila_data.utils.monkey_patch.assert_called_once_with()
        manila_data.service.Service.create.assert_called_once_with(
            binary='manila-data')
        manila_data.service.wait.assert_called_once_with()
        manila_data.service.serve.assert_called_once_with(
            manila_data.service.Service.create.return_value)



# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import six
from webob import exc

from manila.api.openstack import wsgi
from manila.api.views import export_locations as export_locations_views
from manila.db import api as db_api
from manila import exception
from manila.i18n import _


class ShareReplicaExportLocationController(wsgi.Controller):
    """The Share Instance Export Locations API controller."""

    def __init__(self):
        self._view_builder_class = export_locations_views.ViewBuilder
        self.resource_name = 'share_replica_export_location'
        super(ShareReplicaExportLocationController, self).__init__()

    def _verify_share_replica(self, context, share_replica_id):
        try:
            db_api.share_replica_get(context, share_replica_id)
        except exception.NotFound:
            msg = _("Share replica '%s' not found.") % share_replica_id
            raise exc.HTTPNotFound(explanation=msg)

    @wsgi.Controller.api_version('2.47', experimental=True)
    @wsgi.Controller.authorize
    def index(self, req, share_replica_id):
        """Return a list of export locations for the share instance."""
        context = req.environ['manila.context']
        self._verify_share_replica(context, share_replica_id)
        export_locations = (
            db_api.share_export_locations_get_by_share_instance_id(
                context, share_replica_id,
                include_admin_only=context.is_admin)
        )
        return self._view_builder.summary_list(req, export_locations,
                                               replica=True)

    @wsgi.Controller.api_version('2.47', experimental=True)
    @wsgi.Controller.authorize
    def show(self, req, share_replica_id, export_location_uuid):
        """Return data about the requested export location."""
        context = req.environ['manila.context']
        self._verify_share_replica(context, share_replica_id)
        try:
            export_location = db_api.share_export_location_get_by_uuid(
                context, export_location_uuid)
            return self._view_builder.detail(req, export_location,
                                             replica=True)
        except exception.ExportLocationNotFound as e:
            raise exc.HTTPNotFound(explanation=six.text_type(e))


def create_resource():
    return wsgi.Resource(ShareReplicaExportLocationController())



# Copyright (c) 2016, MapR Technologies
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""
Utility for processing MapR cluster operations
"""

import json
import pipes
import socket

from oslo_concurrency import processutils
from oslo_log import log
import six

from manila.common import constants
from manila import exception
from manila.i18n import _
from manila import utils

LOG = log.getLogger(__name__)


def get_version_handler(configuration):
    # here can be choosing DriverUtils depend on cluster version
    return BaseDriverUtil(configuration)


class BaseDriverUtil(object):
    """Utility class for MapR-FS specific operations."""
    NOT_FOUND_MSG = 'No such'
    ERROR_MSG = 'ERROR'

    def __init__(self, configuration):
        self.configuration = configuration
        self.ssh_connections = {}
        self.hosts = self.configuration.maprfs_clinode_ip
        self.local_hosts = socket.gethostbyname_ex(socket.gethostname())[2]
        self.maprcli_bin = '/usr/bin/maprcli'
        self.hadoop_bin = '/usr/bin/hadoop'

    def _execute(self, *cmd, **kwargs):
        for x in range(0, len(self.hosts)):
            try:
                check_exit_code = kwargs.pop('check_exit_code', True)
                host = self.hosts[x]
                if host in self.local_hosts:
                    cmd = self._as_user(cmd,
                                        self.configuration.maprfs_ssh_name)
                    out, err = utils.execute(*cmd,
                                             check_exit_code=check_exit_code)
                else:
                    out, err = self._run_ssh(host, cmd, check_exit_code)
                # move available cldb host to the beginning
                if x > 0:
                    self.hosts[0], self.hosts[x] = self.hosts[x], self.hosts[0]
                return out, err
            except exception.ProcessExecutionError as e:
                if self._check_error(e):
                    raise
                elif x < len(self.hosts) - 1:
                    msg = ('Error running SSH command. Trying another host')
                    LOG.error(msg)
                else:
                    raise
            except Exception as e:
                if x < len(self.hosts) - 1:
                    msg = ('Error running SSH command. Trying another host')
                    LOG.error(msg)
                else:
                    raise exception.ProcessExecutionError(six.text_type(e))

    def _run_ssh(self, host, cmd_list, check_exit_code=False):
        command = ' '.join(pipes.quote(cmd_arg) for cmd_arg in cmd_list)
        connection = self.ssh_connections.get(host)
        if connection is None:
            ssh_name = self.configuration.maprfs_ssh_name
            password = self.configuration.maprfs_ssh_pw
            private_key = self.configuration.maprfs_ssh_private_key
            remote_ssh_port = self.configuration.maprfs_ssh_port
            ssh_conn_timeout = self.configuration.ssh_conn_timeout
            min_size = self.configuration.ssh_min_pool_conn
            max_size = self.configuration.ssh_max_pool_conn

            ssh_pool = utils.SSHPool(host,
                                     remote_ssh_port,
                                     ssh_conn_timeout,
                                     ssh_name,
                                     password=password,
                                     privatekey=private_key,
                                     min_size=min_size,
                                     max_size=max_size)
            ssh = ssh_pool.create()
            self.ssh_connections[host] = (ssh_pool, ssh)
        else:
            ssh_pool, ssh = connection

        if not ssh.get_transport().is_active():
            ssh_pool.remove(ssh)
            ssh = ssh_pool.create()
            self.ssh_connections[host] = (ssh_pool, ssh)
        return processutils.ssh_execute(
            ssh,
            command,
            check_exit_code=check_exit_code)

    @staticmethod
    def _check_error(error):
        # check if error was native
        return BaseDriverUtil.ERROR_MSG in error.stdout

    @staticmethod
    def _as_user(cmd, user):
        return ['sudo', 'su', '-', user, '-c',
                ' '.join(pipes.quote(cmd_arg) for cmd_arg in cmd)]

    @staticmethod
    def _add_params(cmd, **kwargs):
        params = []
        for x in kwargs.keys():
            params.append('-' + x)
            params.append(kwargs[x])
        return cmd + params

    def create_volume(self, name, path, size, **kwargs):
        # delete size param as it is set separately
        if kwargs.get('quota'):
            del kwargs['quota']
        sizestr = six.text_type(size) + 'G'
        cmd = [self.maprcli_bin, 'volume', 'create', '-name',
               name, '-path', path, '-quota',
               sizestr, '-readAce', '', '-writeAce', '']
        cmd = self._add_params(cmd, **kwargs)
        self._execute(*cmd)

    def volume_exists(self, volume_name):
        cmd = [self.maprcli_bin, 'volume', 'info', '-name', volume_name]
        out, __ = self._execute(*cmd, check_exit_code=False)
        return self.NOT_FOUND_MSG not in out

    def delete_volume(self, name):
        cmd = [self.maprcli_bin, 'volume', 'remove', '-name', name, '-force',
               'true']
        out, __ = self._execute(*cmd, check_exit_code=False)
        # if volume does not exist do not raise exception.ProcessExecutionError
        if self.ERROR_MSG in out and self.NOT_FOUND_MSG not in out:
            raise exception.ProcessExecutionError(out)

    def set_volume_size(self, name, size):
        sizestr = six.text_type(size) + 'G'
        cmd = [self.maprcli_bin, 'volume', 'modify', '-name', name, '-quota',
               sizestr]
        self._execute(*cmd)

    def create_snapshot(self, name, volume_name):
        cmd = [self.maprcli_bin, 'volume', 'snapshot', 'create',
               '-snapshotname',
               name, '-volume', volume_name]
        self._execute(*cmd)

    def delete_snapshot(self, name, volume_name):
        cmd = [self.maprcli_bin, 'volume', 'snapshot', 'remove',
               '-snapshotname',
               name, '-volume', volume_name]
        out, __ = self._execute(*cmd, check_exit_code=False)
        # if snapshot does not exist do not raise ProcessExecutionError
        if self.ERROR_MSG in out and self.NOT_FOUND_MSG not in out:
            raise exception.ProcessExecutionError(out)

    def get_volume_info(self, volume_name, columns=None):
        cmd = [self.maprcli_bin, 'volume', 'info', '-name', volume_name,
               '-json']
        if columns:
            cmd += ['-columns', ','.join(columns)]
        out, __ = self._execute(*cmd)
        return json.loads(out)['data'][0]

    def get_volume_info_by_path(self, volume_path, columns=None,
                                check_if_exists=False):
        cmd = [self.maprcli_bin, 'volume', 'info', '-path', volume_path,
               '-json']
        if columns:
            cmd += ['-columns', ','.join(columns)]
        out, __ = self._execute(*cmd, check_exit_code=not check_if_exists)
        if check_if_exists and self.NOT_FOUND_MSG in out:
            return None
        return json.loads(out)['data'][0]

    def get_snapshot_list(self, volume_name=None, volume_path=None):
        params = {}
        if volume_name:
            params['volume'] = volume_name
        if volume_path:
            params['path'] = volume_name
        cmd = [self.maprcli_bin, 'volume', 'snapshot', 'list', '-volume',
               '-columns',
               'snapshotname', '-json']
        cmd = self._add_params(cmd, **params)
        out, __ = self._execute(*cmd)
        return [x['snapshotname'] for x in json.loads(out)['data']]

    def rename_volume(self, name, new_name):
        cmd = [self.maprcli_bin, 'volume', 'rename', '-name', name, '-newname',
               new_name]
        self._execute(*cmd)

    def fs_capacity(self):
        cmd = [self.hadoop_bin, 'fs', '-df']
        out, err = self._execute(*cmd)
        lines = out.splitlines()
        try:
            fields = lines[1].split()
            total = int(fields[1])
            free = int(fields[3])
        except (IndexError, ValueError):
            msg = _('Failed to get MapR-FS capacity info.')
            LOG.exception(msg)
            raise exception.ProcessExecutionError(msg)
        return total, free

    def maprfs_ls(self, path):
        cmd = [self.hadoop_bin, 'fs', '-ls', path]
        out, __ = self._execute(*cmd)
        return out

    def maprfs_cp(self, source, dest):
        cmd = [self.hadoop_bin, 'fs', '-cp', '-p', source, dest]
        self._execute(*cmd)

    def maprfs_chmod(self, dest, mod):
        cmd = [self.hadoop_bin, 'fs', '-chmod', mod, dest]
        self._execute(*cmd)

    def maprfs_du(self, path):
        cmd = [self.hadoop_bin, 'fs', '-du', '-s', path]
        out, __ = self._execute(*cmd)
        return int(out.split(' ')[0])

    def check_state(self):
        cmd = [self.hadoop_bin, 'fs', '-ls', '/']
        out, __ = self._execute(*cmd, check_exit_code=False)
        return 'Found' in out

    def dir_not_empty(self, path):
        cmd = [self.hadoop_bin, 'fs', '-ls', path]
        out, __ = self._execute(*cmd, check_exit_code=False)
        return 'Found' in out

    def set_volume_ace(self, volume_name, access_rules):
        read_accesses = []
        write_accesses = []
        for access_rule in access_rules:
            if access_rule['access_level'] == constants.ACCESS_LEVEL_RO:
                read_accesses.append(access_rule['access_to'])
            elif access_rule['access_level'] == constants.ACCESS_LEVEL_RW:
                read_accesses.append(access_rule['access_to'])
                write_accesses.append(access_rule['access_to'])

        def rule_type(access_to):
            if self.group_exists(access_to):
                return 'g'
            elif self.user_exists(access_to):
                return 'u'
            else:
                # if nor user nor group exits, it should try add group rule
                return 'g'

        read_accesses_string = '|'.join(
            map(lambda x: rule_type(x) + ':' + x, read_accesses))
        write_accesses_string = '|'.join(
            map(lambda x: rule_type(x) + ':' + x, write_accesses))
        cmd = [self.maprcli_bin, 'volume', 'modify', '-name', volume_name,
               '-readAce', read_accesses_string, '-writeAce',
               write_accesses_string]
        self._execute(*cmd)

    def add_volume_ace_rules(self, volume_name, access_rules):
        if not access_rules:
            return
        access_rules_map = self.get_access_rules(volume_name)
        for access_rule in access_rules:
            access_rules_map[access_rule['access_to']] = access_rule
        self.set_volume_ace(volume_name, access_rules_map.values())

    def remove_volume_ace_rules(self, volume_name, access_rules):
        if not access_rules:
            return
        access_rules_map = self.get_access_rules(volume_name)
        for access_rule in access_rules:
            if access_rules_map.get(access_rule['access_to']):
                del access_rules_map[access_rule['access_to']]
        self.set_volume_ace(volume_name, access_rules_map.values())

    def get_access_rules(self, volume_name):
        info = self.get_volume_info(volume_name)
        aces = info['volumeAces']
        read_ace = aces['readAce']
        write_ace = aces['writeAce']
        access_rules_map = {}
        self._retrieve_access_rules_from_ace(read_ace, 'r', access_rules_map)
        self._retrieve_access_rules_from_ace(write_ace, 'w', access_rules_map)
        return access_rules_map

    def _retrieve_access_rules_from_ace(self, ace, ace_type, access_rules_map):
        access = constants.ACCESS_LEVEL_RW if ace_type == 'w' else (
            constants.ACCESS_LEVEL_RO)
        if ace not in ['p', '']:
            write_rules = [x.strip() for x in ace.split('|')]
            for user in write_rules:
                rule_type, username = user.split(':')
                if rule_type not in ['u', 'g']:
                    continue
                access_rules_map[username] = {
                    'access_level': access,
                    'access_to': username,
                    'access_type': 'user',
                }

    def user_exists(self, user):
        cmd = ['getent', 'passwd', user]
        out, __ = self._execute(*cmd, check_exit_code=False)
        return out != ''

    def group_exists(self, group):
        cmd = ['getent', 'group', group]
        out, __ = self._execute(*cmd, check_exit_code=False)
        return out != ''

    def get_cluster_name(self):
        cmd = [self.maprcli_bin, 'dashboard', 'info', '-json']
        out, __ = self._execute(*cmd)
        try:
            return json.loads(out)['data'][0]['cluster']['name']
        except (IndexError, ValueError) as e:
            msg = (_("Failed to parse cluster name. Error: %s") % e)
            raise exception.ProcessExecutionError(msg)



# Copyright 2013 OpenStack LLC.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from manila.api import common
from manila.common import constants


class ViewBuilder(common.ViewBuilder):
    """Model a server API response as a python dictionary."""

    _collection_name = 'security_services'
    _detail_version_modifiers = [
        'add_ou_to_security_service',
    ]

    def summary_list(self, request, security_services):
        """Show a list of security services without many details."""
        return self._list_view(self.summary, request, security_services)

    def detail_list(self, request, security_services):
        """Detailed view of a list of security services."""
        return self._list_view(self.detail, request, security_services)

    def summary(self, request, security_service):
        """Generic, non-detailed view of a security service."""
        return {
            'security_service': {
                'id': security_service.get('id'),
                'name': security_service.get('name'),
                'type': security_service.get('type'),
                # NOTE(vponomaryov): attr "status" was removed from model and
                # is left in view for compatibility purposes since it affects
                # user-facing API. This should be removed right after no one
                # uses it anymore.
                'status': constants.STATUS_NEW,
            }
        }

    def detail(self, request, security_service):
        """Detailed view of a single security service."""
        view = self.summary(request, security_service)
        keys = (
            'created_at', 'updated_at', 'description', 'dns_ip', 'server',
            'domain', 'user', 'password', 'project_id')
        for key in keys:
            view['security_service'][key] = security_service.get(key)
        self.update_versioned_resource_dict(
            request, view['security_service'], security_service)
        return view

    @common.ViewBuilder.versioned_method("2.44")
    def add_ou_to_security_service(self, context, ss_dict, ss):
        ss_dict['ou'] = ss.get('ou')

    def _list_view(self, func, request, security_services):
        """Provide a view for a list of security services."""
        security_services_list = [func(request, service)['security_service']
                                  for service in security_services]
        security_services_dict = dict(security_services=security_services_list)
        return security_services_dict



# Copyright (c) 2015 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Common GlussterFS routines."""


import re
import xml.etree.cElementTree as etree

from oslo_config import cfg
from oslo_log import log
import six

from manila import exception
from manila.i18n import _
from manila.share.drivers.ganesha import utils as ganesha_utils

LOG = log.getLogger(__name__)


glusterfs_common_opts = [
    cfg.StrOpt('glusterfs_server_password',
               secret=True,
               deprecated_name='glusterfs_native_server_password',
               help='Remote GlusterFS server node\'s login password. '
                    'This is not required if '
                    '\'glusterfs_path_to_private_key\' is '
                    'configured.'),
    cfg.StrOpt('glusterfs_path_to_private_key',
               deprecated_name='glusterfs_native_path_to_private_key',
               help='Path of Manila host\'s private SSH key file.'),
]


CONF = cfg.CONF
CONF.register_opts(glusterfs_common_opts)


def _check_volume_presence(f):

    def wrapper(self, *args, **kwargs):
        if not self.components.get('volume'):
            raise exception.GlusterfsException(
                _("Gluster address does not have a volume component."))
        return f(self, *args, **kwargs)

    return wrapper


def volxml_get(xmlout, *paths, **kwargs):
    """Attempt to extract a value by a set of Xpaths from XML."""
    for path in paths:
        value = xmlout.find(path)
        if value is not None:
            break
    if value is None:
        if 'default' in kwargs:
            return kwargs['default']
        raise exception.InvalidShare(
            _("Volume query response XML has no value for any of "
              "the following Xpaths: %s") % ", ".join(paths))
    return value.text


class GlusterManager(object):
    """Interface with a GlusterFS volume."""

    scheme = re.compile('\A(?:(?P<user>[^:@/]+)@)?'
                        '(?P<host>[^:@/]+)'
                        '(?::/(?P<volume>[^/]+)(?P<path>/.*)?)?\Z')

    # See this about GlusterFS' convention for Boolean interpretation
    # of strings:
    # https://github.com/gluster/glusterfs/blob/v3.7.8/
    #         libglusterfs/src/common-utils.c#L1680-L1708
    GLUSTERFS_TRUE_VALUES = ('ON', 'YES', 'TRUE', 'ENABLE', '1')
    GLUSTERFS_FALSE_VALUES = ('OFF', 'NO', 'FALSE', 'DISABLE', '0')

    @classmethod
    def parse(cls, address):
        """Parse address string into component dict."""
        m = cls.scheme.search(address)
        if not m:
            raise exception.GlusterfsException(
                _('Invalid gluster address %s.') % address)
        return m.groupdict()

    def __getattr__(self, attr):
        if attr in self.components:
            return self.components[attr]
        raise AttributeError("'%(typ)s' object has no attribute '%(attr)s'" %
                             {'typ': type(self).__name__, 'attr': attr})

    def __init__(self, address, execf=None, path_to_private_key=None,
                 remote_server_password=None, requires={}):
        """Initialize a GlusterManager instance.

        :param address: the Gluster URI (either string of
                        [<user>@]<host>[:/<volume>[/<path>]] format or
                        component dict with "user", "host", "volume",
                        "path" keys).
        :param execf: executor function for management commands.
        :param path_to_private_key: path to private ssh key of remote server.
        :param remote_server_password: ssh password for remote server.
        :param requires: a dict mapping some of the component names to
                         either True or False; having it specified,
                         respectively, the presence or absence of the
                         given component in the uri will be enforced.
        """

        if isinstance(address, dict):
            tmp_addr = ""
            if address.get('user') is not None:
                tmp_addr = address.get('user') + '@'
            if address.get('host') is not None:
                tmp_addr += address.get('host')
            if address.get('volume') is not None:
                tmp_addr += ':/' + address.get('volume')
            if address.get('path') is not None:
                tmp_addr += address.get('path')
            self.components = self.parse(tmp_addr)
            # Verify that the original dictionary matches the parsed
            # dictionary. This will flag typos such as {'volume': 'vol/err'}
            # in the original dictionary as errors.  Additionally,
            # extra keys will need to be flagged as an error.
            sanitized_address = {key: None for key in self.scheme.groupindex}
            sanitized_address.update(address)
            if sanitized_address != self.components:
                raise exception.GlusterfsException(
                    _('Invalid gluster address %s.') % address)
        else:
            self.components = self.parse(address)

        for k, v in requires.items():
            if v is None:
                continue
            if (self.components.get(k) is not None) != v:
                raise exception.GlusterfsException(
                    _('Invalid gluster address %s.') % address)

        self.path_to_private_key = path_to_private_key
        self.remote_server_password = remote_server_password
        if execf:
            self.gluster_call = self.make_gluster_call(execf)

    @property
    def host_access(self):
        return '@'.join(filter(None, (self.user, self.host)))

    def _build_uri(self, base):
        u = base
        for sep, comp in ((':/', 'volume'), ('', 'path')):
            if self.components[comp] is None:
                break
            u = sep.join((u, self.components[comp]))
        return u

    @property
    def qualified(self):
        return self._build_uri(self.host_access)

    @property
    def export(self):
        if self.volume:
            return self._build_uri(self.host)

    def make_gluster_call(self, execf):
        """Execute a Gluster command locally or remotely."""
        if self.user:
            gluster_execf = ganesha_utils.SSHExecutor(
                self.host, 22, None, self.user,
                password=self.remote_server_password,
                privatekey=self.path_to_private_key)
        else:
            gluster_execf = ganesha_utils.RootExecutor(execf)

        def _gluster_call(*args, **kwargs):
            logmsg = kwargs.pop('log', None)
            error_policy = kwargs.pop('error_policy', 'coerce')
            if (error_policy not in ('raw', 'coerce', 'suppress') and
               not isinstance(error_policy[0], int)):
                raise TypeError(_("undefined error_policy %s") %
                                repr(error_policy))

            try:
                return gluster_execf(*(('gluster',) + args), **kwargs)
            except exception.ProcessExecutionError as exc:
                if error_policy == 'raw':
                    raise
                elif error_policy == 'coerce':
                    pass
                elif (error_policy == 'suppress' or
                      exc.exit_code in error_policy):
                    return
                if logmsg:
                    LOG.error("%s: GlusterFS instrumentation failed.",
                              logmsg)
                raise exception.GlusterfsException(
                    _("GlusterFS management command '%(cmd)s' failed "
                      "with details as follows:\n%(details)s.") % {
                        'cmd': ' '.join(args),
                        'details': exc})

        return _gluster_call

    def xml_response_check(self, xmlout, command, countpath=None):
        """Sanity check for GlusterFS XML response."""
        commandstr = ' '.join(command)
        ret = {}
        for e in 'opRet', 'opErrno':
            ret[e] = int(volxml_get(xmlout, e))
        if ret == {'opRet': -1, 'opErrno': 0}:
            raise exception.GlusterfsException(_(
                'GlusterFS command %(command)s on volume %(volume)s failed'
            ) % {'volume': self.volume, 'command': command})
        if list(ret.values()) != [0, 0]:
            errdct = {'volume': self.volume, 'command': commandstr,
                      'opErrstr': volxml_get(xmlout, 'opErrstr', default=None)}
            errdct.update(ret)
            raise exception.InvalidShare(_(
                'GlusterFS command %(command)s on volume %(volume)s got '
                'unexpected response: '
                'opRet=%(opRet)s, opErrno=%(opErrno)s, opErrstr=%(opErrstr)s'
            ) % errdct)
        if not countpath:
            return
        count = volxml_get(xmlout, countpath)
        if count != '1':
            raise exception.InvalidShare(
                _('GlusterFS command %(command)s on volume %(volume)s got '
                  'ambiguous response: '
                  '%(count)s records') % {
                    'volume': self.volume, 'command': commandstr,
                    'count': count})

    def _get_vol_option_via_info(self, option):
        """Get the value of an option set on a GlusterFS volume via volinfo."""
        args = ('--xml', 'volume', 'info', self.volume)
        out, err = self.gluster_call(*args, log=("retrieving volume info"))

        if not out:
            raise exception.GlusterfsException(
                'gluster volume info %s: no data received' %
                self.volume
            )

        volxml = etree.fromstring(out)
        self.xml_response_check(volxml, args[1:], './volInfo/volumes/count')
        for e in volxml.findall(".//option"):
            o, v = (volxml_get(e, a) for a in ('name', 'value'))
            if o == option:
                return v

    @_check_volume_presence
    def _get_vol_user_option(self, useropt):
        """Get the value of an user option set on a GlusterFS volume."""
        option = '.'.join(('user', useropt))
        return self._get_vol_option_via_info(option)

    @_check_volume_presence
    def _get_vol_regular_option(self, option):
        """Get the value of a regular option set on a GlusterFS volume."""
        args = ('--xml', 'volume', 'get', self.volume, option)

        out, err = self.gluster_call(*args, check_exit_code=False)

        if not out:
            # all input is valid, but the option has not been set
            # (nb. some options do come by a null value, but some
            # don't even have that, see eg. cluster.nufa)
            return

        try:
            optxml = etree.fromstring(out)
        except Exception:
            # non-xml output indicates that GlusterFS backend does not support
            # 'vol get', we fall back to 'vol info' based retrieval (glusterfs
            # < 3.7).
            return self._get_vol_option_via_info(option)

        self.xml_response_check(optxml, args[1:], './volGetopts/count')
        # the Xpath has changed from first to second as of GlusterFS
        # 3.7.14 (see http://review.gluster.org/14931).
        return volxml_get(optxml, './volGetopts/Value',
                          './volGetopts/Opt/Value')

    def get_vol_option(self, option, boolean=False):
        """Get the value of an option set on a GlusterFS volume."""
        useropt = re.sub('\Auser\.', '', option)
        if option == useropt:
            value = self._get_vol_regular_option(option)
        else:
            value = self._get_vol_user_option(useropt)
        if not boolean or value is None:
            return value
        if value.upper() in self.GLUSTERFS_TRUE_VALUES:
            return True
        if value.upper() in self.GLUSTERFS_FALSE_VALUES:
            return False
        raise exception.GlusterfsException(_(
            "GlusterFS volume option on volume %(volume)s: "
            "%(option)s=%(value)s cannot be interpreted as Boolean") % {
                'volume': self.volume, 'option': option, 'value': value})

    @_check_volume_presence
    def set_vol_option(self, option, value, ignore_failure=False):
        value = {True: self.GLUSTERFS_TRUE_VALUES[0],
                 False: self.GLUSTERFS_FALSE_VALUES[0]}.get(value, value)
        if value is None:
            args = ('reset', (option,))
        else:
            args = ('set', (option, value))
        policy = (1,) if ignore_failure else 'coerce'
        self.gluster_call(
            'volume', args[0], self.volume, *args[1], error_policy=policy)

    def get_gluster_version(self):
        """Retrieve GlusterFS version.

        :returns: version (as tuple of strings, example: ('3', '6', '0beta2'))
        """
        out, err = self.gluster_call('--version',
                                     log=("GlusterFS version query"))
        try:
            owords = out.split()
            if owords[0] != 'glusterfs':
                raise RuntimeError
            vers = owords[1].split('.')
            # provoke an exception if vers does not start with two numerals
            int(vers[0])
            int(vers[1])
        except Exception:
            raise exception.GlusterfsException(
                _("Cannot parse version info obtained from server "
                  "%(server)s, version info: %(info)s") %
                {'server': self.host, 'info': out})
        return vers

    def check_gluster_version(self, minvers):
        """Retrieve and check GlusterFS version.

        :param minvers: minimum version to require
                        (given as tuple of integers, example: (3, 6))
        """
        vers = self.get_gluster_version()
        if numreduct(vers) < minvers:
            raise exception.GlusterfsException(_(
                "Unsupported GlusterFS version %(version)s on server "
                "%(server)s, minimum requirement: %(minvers)s") % {
                'server': self.host,
                'version': '.'.join(vers),
                'minvers': '.'.join(six.text_type(c) for c in minvers)})


def numreduct(vers):
    """The numeric reduct of a tuple of strings.

    That is, applying an integer conversion map on the longest
    initial segment of vers which consists of numerals.
    """
    numvers = []
    for c in vers:
        try:
            numvers.append(int(c))
        except ValueError:
            break
    return tuple(numvers)


def _mount_gluster_vol(execute, gluster_export, mount_path, ensure=False):
    """Mount a GlusterFS volume at the specified mount path.

    :param execute: command execution function
    :param gluster_export: GlusterFS export to mount
    :param mount_path: path to mount at
    :param ensure: boolean to allow remounting a volume with a warning
    """
    execute('mkdir', '-p', mount_path)
    command = ['mount', '-t', 'glusterfs', gluster_export, mount_path]
    try:
        execute(*command, run_as_root=True)
    except exception.ProcessExecutionError as exc:
        if ensure and 'already mounted' in exc.stderr:
            LOG.warning("%s is already mounted.", gluster_export)
        else:
            raise exception.GlusterfsException(
                'Unable to mount Gluster volume'
            )


def _umount_gluster_vol(execute, mount_path):
    """Unmount a GlusterFS volume at the specified mount path.

    :param execute: command execution function
    :param mount_path: path where volume is mounted
    """

    try:
        execute('umount', mount_path, run_as_root=True)
    except exception.ProcessExecutionError as exc:
        msg = (_("Unable to unmount gluster volume. "
                 "mount_dir: %(mount_path)s, Error: %(error)s") %
               {'mount_path': mount_path, 'error': exc.stderr})
        LOG.error(msg)
        raise exception.GlusterfsException(msg)


def _restart_gluster_vol(gluster_mgr):
    """Restart a GlusterFS volume through its manager.

    :param gluster_mgr: GlusterManager instance
    """

    # TODO(csaba): '--mode=script' ensures that the Gluster CLI runs in
    # script mode. This seems unnecessary as the Gluster CLI is
    # expected to run in non-interactive mode when the stdin is not
    # a terminal, as is the case below. But on testing, found the
    # behaviour of Gluster-CLI to be the contrary. Need to investigate
    # this odd-behaviour of Gluster-CLI.
    gluster_mgr.gluster_call(
        'volume', 'stop', gluster_mgr.volume, '--mode=script',
        log=("stopping GlusterFS volume %s") % gluster_mgr.volume)

    gluster_mgr.gluster_call(
        'volume', 'start', gluster_mgr.volume,
        log=("starting GlusterFS volume %s") % gluster_mgr.volume)



# Copyright (c) 2015 Huawei Technologies Co., Ltd.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Abstract base class to work with share."""
import abc

import six


@six.add_metaclass(abc.ABCMeta)
class HuaweiBase(object):
    """Interface to work with share."""

    def __init__(self, configuration):
        """Do initialization."""
        self.configuration = configuration

    @abc.abstractmethod
    def create_share(self, share, share_server):
        """Is called to create share."""

    @abc.abstractmethod
    def create_snapshot(self, snapshot, share_server):
        """Is called to create snapshot."""

    @abc.abstractmethod
    def delete_share(self, share, share_server):
        """Is called to remove share."""

    @abc.abstractmethod
    def delete_snapshot(self, snapshot, share_server):
        """Is called to remove snapshot."""

    @abc.abstractmethod
    def allow_access(self, share, access, share_server):
        """Allow access to the share."""

    @abc.abstractmethod
    def deny_access(self, share, access, share_server):
        """Deny access to the share."""

    @abc.abstractmethod
    def ensure_share(self, share, share_server=None):
        """Ensure that share is exported."""

    @abc.abstractmethod
    def update_access(self, share, access_rules, add_rules,
                      delete_rules, share_server):
        """Update access rules list."""

    @abc.abstractmethod
    def extend_share(self, share, new_size, share_server):
        """Extends size of existing share."""

    @abc.abstractmethod
    def create_share_from_snapshot(self, share, snapshot,
                                   share_server=None):
        """Create share from snapshot."""

    @abc.abstractmethod
    def shrink_share(self, share, new_size, share_server):
        """Shrinks size of existing share."""

    @abc.abstractmethod
    def manage_existing(self, share, driver_options):
        """Manage existing share."""

    @abc.abstractmethod
    def manage_existing_snapshot(self, snapshot, driver_options):
        """Manage existing snapshot."""

    @abc.abstractmethod
    def get_network_allocations_number(self):
        """Get number of network interfaces to be created."""

    @abc.abstractmethod
    def get_pool(self, share):
        """Return pool name where the share resides on."""

    def update_share_stats(self, stats_dict):
        """Retrieve stats info from share group."""

    @abc.abstractmethod
    def setup_server(self, network_info, metadata=None):
        """Set up share server with given network parameters."""

    @abc.abstractmethod
    def teardown_server(self, server_details, security_services=None):
        """Teardown share server."""

    @abc.abstractmethod
    def create_replica(self, context, replica_list, new_replica,
                       access_rules, replica_snapshots, share_server=None):
        """Replicate the active replica to a new replica on this backend."""

    @abc.abstractmethod
    def update_replica_state(self, context, replica_list, replica,
                             access_rules, replica_snapshots,
                             share_server=None):
        """Update the replica_state of a replica."""

    @abc.abstractmethod
    def promote_replica(self, context, replica_list, replica, access_rules,
                        share_server=None):
        """Promote a replica to 'active' replica state."""

    @abc.abstractmethod
    def delete_replica(self, context, replica_list, replica_snapshots,
                       replica, share_server=None):
        """Delete a replica."""

    @abc.abstractmethod
    def revert_to_snapshot(self, context, snapshot, share_access_rules,
                           snapshot_access_rules, share_server=None):
        """Revert a snapshot."""



# Copyright 2016 Mirantis, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""Unit tests for the Protocol helper module."""

import ddt
import functools
import mock

from manila.common import constants as const
from manila import exception
from manila.share.drivers.container import protocol_helper
from manila import test
from manila.tests.share.drivers.container.fakes import fake_share


@ddt.ddt
class DockerCIFSHelperTestCase(test.TestCase):
    """Tests ContainerShareDriver"""

    def setUp(self):
        super(DockerCIFSHelperTestCase, self).setUp()
        self._helper = mock.Mock()
        self.fake_conf = mock.Mock()
        self.fake_conf.container_cifs_guest_ok = "yes"
        self.DockerCIFSHelper = protocol_helper.DockerCIFSHelper(
            self._helper, share=fake_share(), config=self.fake_conf)

    def fake_exec_sync(self, *args, **kwargs):
        kwargs["execute_arguments"].append(args)
        try:
            ret_val = kwargs["ret_val"]
        except KeyError:
            ret_val = None
        return [ret_val]

    def test_create_share_guest_ok(self):
        expected_arguments = [
            ("fakeserver", ["net", "conf", "addshare", "fakeshareid",
             "/shares/fakeshareid", "writeable=y", "guest_ok=y"]),
            ("fakeserver", ["net", "conf", "setparm", "fakeshareid",
             "browseable", "yes"]),
            ("fakeserver", ["net", "conf", "setparm", "fakeshareid",
             "hosts allow", "127.0.0.1"]),
            ("fakeserver", ["net", "conf", "setparm", "fakeshareid",
             "read only", "no"]),
            ("fakeserver", ["net", "conf", "setparm", "fakeshareid",
             "hosts deny", "0.0.0.0/0"]),
            ("fakeserver", ["net", "conf", "setparm", "fakeshareid",
             "create mask", "0755"])]
        actual_arguments = []
        self._helper.execute = functools.partial(
            self.fake_exec_sync, execute_arguments=actual_arguments,
            ret_val=" fake 192.0.2.2/24 more fake \n" * 20)
        self.DockerCIFSHelper.share = fake_share()

        self.DockerCIFSHelper.create_share("fakeserver")

        self.assertEqual(expected_arguments.sort(), actual_arguments.sort())

    def test_create_share_guest_not_ok(self):
        self.DockerCIFSHelper.conf = mock.Mock()
        self.DockerCIFSHelper.conf.container_cifs_guest_ok = False
        expected_arguments = [
            ("fakeserver", ["net", "conf", "addshare", "fakeshareid",
             "/shares/fakeshareid", "writeable=y", "guest_ok=n"]),
            ("fakeserver", ["net", "conf", "setparm", "fakeshareid",
             "browseable", "yes"]),
            ("fakeserver", ["net", "conf", "setparm", "fakeshareid",
             "hosts allow", "192.0.2.2"]),
            ("fakeserver", ["net", "conf", "setparm", "fakeshareid",
             "read only", "no"]),
            ("fakeserver", ["net", "conf", "setparm", "fakeshareid",
             "hosts deny", "0.0.0.0/0"]),
            ("fakeserver", ["net", "conf", "setparm", "fakeshareid",
             "create mask", "0755"])]
        actual_arguments = []
        self._helper.execute = functools.partial(
            self.fake_exec_sync, execute_arguments=actual_arguments,
            ret_val=" fake 192.0.2.2/24 more fake \n" * 20)
        self.DockerCIFSHelper.share = fake_share()

        self.DockerCIFSHelper.create_share("fakeserver")

        self.assertEqual(expected_arguments.sort(), actual_arguments.sort())

    def test_delete_share(self):
        self.DockerCIFSHelper.share = fake_share()

        self.DockerCIFSHelper.delete_share("fakeserver", "fakeshareid")

        self.DockerCIFSHelper.container.execute.assert_called_with(
            "fakeserver",
            ["net", "conf", "delshare", "fakeshareid"], ignore_errors=False)

    def test__get_access_group_ro(self):
        result = self.DockerCIFSHelper._get_access_group(const.ACCESS_LEVEL_RO)

        self.assertEqual("read list", result)

    def test__get_access_group_rw(self):
        result = self.DockerCIFSHelper._get_access_group(const.ACCESS_LEVEL_RW)

        self.assertEqual("valid users", result)

    def test__get_access_group_other(self):
        self.assertRaises(exception.InvalidShareAccessLevel,
                          self.DockerCIFSHelper._get_access_group,
                          "fake_level")

    def test__get_existing_users(self):
        self.DockerCIFSHelper.container.execute = mock.Mock(
            return_value=("fake_user", ""))

        result = self.DockerCIFSHelper._get_existing_users("fake_server_id",
                                                           "fake_share",
                                                           "fake_access")

        self.assertEqual("fake_user", result)
        self.DockerCIFSHelper.container.execute.assert_called_once_with(
            "fake_server_id",
            ["net", "conf", "getparm", "fake_share", "fake_access"],
            ignore_errors=True)

    def test__set_users(self):
        self.DockerCIFSHelper.container.execute = mock.Mock()

        self.DockerCIFSHelper._set_users("fake_server_id", "fake_share",
                                         "fake_access", "fake_user")

        self.DockerCIFSHelper.container.execute.assert_called_once_with(
            "fake_server_id",
            ["net", "conf", "setparm", "fake_share", "fake_access",
             "fake_user"])

    def test__allow_access_ok(self):
        self.DockerCIFSHelper._get_access_group = mock.Mock(
            return_value="valid users")
        self.DockerCIFSHelper._get_existing_users = mock.Mock(
            return_value="fake_user")
        self.DockerCIFSHelper._set_users = mock.Mock()

        self.DockerCIFSHelper._allow_access("fake_share", "fake_server_id",
                                            "fake_user2", "rw")

        self.DockerCIFSHelper._get_access_group.assert_called_once_with("rw")
        self.DockerCIFSHelper._get_existing_users.assert_called_once_with(
            "fake_server_id", "fake_share", "valid users")
        self.DockerCIFSHelper._set_users.assert_called_once_with(
            "fake_server_id", "fake_share", "valid users",
            "fake_user fake_user2")

    def test__allow_access_not_ok(self):
        self.DockerCIFSHelper._get_access_group = mock.Mock(
            return_value="valid users")
        self.DockerCIFSHelper._get_existing_users = mock.Mock()
        self.DockerCIFSHelper._get_existing_users.side_effect = TypeError
        self.DockerCIFSHelper._set_users = mock.Mock()

        self.DockerCIFSHelper._allow_access("fake_share", "fake_server_id",
                                            "fake_user2", "rw")

        self.DockerCIFSHelper._get_access_group.assert_called_once_with("rw")
        self.DockerCIFSHelper._get_existing_users.assert_called_once_with(
            "fake_server_id", "fake_share", "valid users")
        self.DockerCIFSHelper._set_users.assert_called_once_with(
            "fake_server_id", "fake_share", "valid users",
            "fake_user2")

    def test__deny_access_ok(self):
        self.DockerCIFSHelper._get_access_group = mock.Mock(
            return_value="valid users")
        self.DockerCIFSHelper._get_existing_users = mock.Mock(
            return_value="fake_user fake_user2")
        self.DockerCIFSHelper._set_users = mock.Mock()

        self.DockerCIFSHelper._deny_access("fake_share", "fake_server_id",
                                           "fake_user2", "rw")

        self.DockerCIFSHelper._get_access_group.assert_called_once_with("rw")
        self.DockerCIFSHelper._get_existing_users.assert_called_once_with(
            "fake_server_id", "fake_share", "valid users")
        self.DockerCIFSHelper._set_users.assert_called_once_with(
            "fake_server_id", "fake_share", "valid users",
            "fake_user")

    def test__deny_access_ok_so_many_users(self):
        self.DockerCIFSHelper._get_access_group = mock.Mock(
            return_value="valid users")
        self.DockerCIFSHelper._get_existing_users = mock.Mock(
            return_value="joost jaap huub dirk")
        self.DockerCIFSHelper._set_users = mock.Mock()

        # Sorry, Jaap.
        self.DockerCIFSHelper._deny_access("fake_share", "fake_server_id",
                                           "jaap", "rw")

        self.DockerCIFSHelper._get_access_group.assert_called_once_with("rw")
        self.DockerCIFSHelper._get_existing_users.assert_called_once_with(
            "fake_server_id", "fake_share", "valid users")
        self.DockerCIFSHelper._set_users.assert_called_once_with(
            "fake_server_id", "fake_share", "valid users",
            "dirk huub joost")

    def test__deny_access_not_ok(self):
        self.DockerCIFSHelper._get_access_group = mock.Mock(
            return_value="valid users")
        self.DockerCIFSHelper._get_existing_users = mock.Mock()
        self.DockerCIFSHelper._get_existing_users.side_effect = TypeError
        self.DockerCIFSHelper._set_users = mock.Mock()
        self.mock_object(protocol_helper.LOG, "warning")

        self.DockerCIFSHelper._deny_access("fake_share", "fake_server_id",
                                           "fake_user2", "rw")

        self.DockerCIFSHelper._get_access_group.assert_called_once_with("rw")
        self.DockerCIFSHelper._get_existing_users.assert_called_once_with(
            "fake_server_id", "fake_share", "valid users")
        self.assertFalse(self.DockerCIFSHelper._set_users.called)
        self.assertTrue(protocol_helper.LOG.warning.called)

    def test_update_access_access_rules_wrong_type(self):
        allow_rules = [{
            "access_to": "192.0.2.2",
            "access_level": "ro",
            "access_type": "fake"
        }]
        self.mock_object(self.DockerCIFSHelper, "_allow_access")

        self.assertRaises(exception.InvalidShareAccess,
                          self.DockerCIFSHelper.update_access,
                          "fakeserver", "fakeshareid", allow_rules, [], [])

    def test_update_access_access_rules_ok(self):
        access_rules = [{
            "access_to": "fakeuser",
            "access_level": "ro",
            "access_type": "user"
        }]
        self.mock_object(self.DockerCIFSHelper, "_allow_access")
        self.DockerCIFSHelper.container.execute = mock.Mock()

        self.DockerCIFSHelper.update_access("fakeserver", "fakeshareid",
                                            access_rules, [], [])

        self.DockerCIFSHelper._allow_access.assert_called_once_with(
            "fakeshareid",
            "fakeserver",
            "fakeuser",
            "ro")
        self.DockerCIFSHelper.container.execute.assert_called_once_with(
            "fakeserver",
            ["net", "conf", "setparm", "fakeshareid", "valid users", ""])

    def test_update_access_add_rules(self):
        add_rules = [{
            "access_to": "fakeuser",
            "access_level": "ro",
            "access_type": "user"
        }]
        self.mock_object(self.DockerCIFSHelper, "_allow_access")

        self.DockerCIFSHelper.update_access("fakeserver", "fakeshareid", [],
                                            add_rules, [])

        self.DockerCIFSHelper._allow_access.assert_called_once_with(
            "fakeshareid",
            "fakeserver",
            "fakeuser",
            "ro")

    def test_update_access_delete_rules(self):
        delete_rules = [{
            "access_to": "fakeuser",
            "access_level": "ro",
            "access_type": "user"
        }]
        self.mock_object(self.DockerCIFSHelper, "_deny_access")

        self.DockerCIFSHelper.update_access("fakeserver", "fakeshareid",
                                            [], [], delete_rules)

        self.DockerCIFSHelper._deny_access.assert_called_once_with(
            "fakeshareid",
            "fakeserver",
            "fakeuser",
            "ro")



# Copyright (c) 2016 Dell Inc. or its subsidiaries.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import pipes

from oslo_concurrency import processutils
from oslo_log import log
from oslo_utils import excutils
import six
from six.moves import http_cookiejar
from six.moves.urllib import error as url_error
from six.moves.urllib import request as url_request

from manila import exception
from manila.i18n import _
from manila.share.drivers.dell_emc.common.enas import constants
from manila.share.drivers.dell_emc.common.enas import utils as enas_utils
from manila import utils

LOG = log.getLogger(__name__)


class XMLAPIConnector(object):
    def __init__(self, configuration, debug=True):
        super(XMLAPIConnector, self).__init__()
        self.storage_ip = enas_utils.convert_ipv6_format_if_needed(
            configuration.emc_nas_server)
        self.username = configuration.emc_nas_login
        self.password = configuration.emc_nas_password
        self.debug = debug
        self.auth_url = 'https://' + self.storage_ip + '/Login'
        self._url = 'https://{}/servlets/CelerraManagementServices'.format(
            self.storage_ip)
        context = enas_utils.create_ssl_context(configuration)
        if context:
            https_handler = url_request.HTTPSHandler(context=context)
        else:
            https_handler = url_request.HTTPSHandler()
        cookie_handler = url_request.HTTPCookieProcessor(
            http_cookiejar.CookieJar())
        self.url_opener = url_request.build_opener(https_handler,
                                                   cookie_handler)
        self._do_setup()

    def _do_setup(self):
        credential = ('user=' + self.username
                      + '&password=' + self.password
                      + '&Login=Login')
        req = url_request.Request(self.auth_url, credential.encode(),
                                  constants.CONTENT_TYPE_URLENCODE)
        resp = self.url_opener.open(req)
        resp_body = resp.read()
        self._http_log_resp(resp, resp_body)

    def _http_log_req(self, req):
        if not self.debug:
            return

        string_parts = ['curl -i']
        string_parts.append(' -X %s' % req.get_method())

        for k in req.headers:
            header = ' -H "%s: %s"' % (k, req.headers[k])
            string_parts.append(header)

        if req.data:
            string_parts.append(" -d '%s'" % req.data)
        string_parts.append(' ' + req.get_full_url())
        LOG.debug("\nREQ: %s.\n", "".join(string_parts))

    def _http_log_resp(self, resp, body):
        if not self.debug:
            return

        headers = six.text_type(resp.headers).replace('\n', '\\n')

        LOG.debug(
            'RESP: [%(code)s] %(resp_hdrs)s\n'
            'RESP BODY: %(resp_b)s.\n',
            {
                'code': resp.getcode(),
                'resp_hdrs': headers,
                'resp_b': body,
            }
        )

    def _request(self, req_body=None, method=None,
                 header=constants.CONTENT_TYPE_URLENCODE):
        req = url_request.Request(self._url, req_body.encode(), header)
        if method not in (None, 'GET', 'POST'):
            req.get_method = lambda: method
        self._http_log_req(req)
        try:
            resp = self.url_opener.open(req)
            resp_body = resp.read()
            self._http_log_resp(resp, resp_body)
        except url_error.HTTPError as http_err:
            if '403' == six.text_type(http_err.code):
                raise exception.NotAuthorized()
            else:
                err = {'errorCode': -1,
                       'httpStatusCode': http_err.code,
                       'messages': six.text_type(http_err),
                       'request': req_body}
                msg = (_("The request is invalid. Reason: %(reason)s") %
                       {'reason': err})
                raise exception.ManilaException(message=msg)

        return resp_body

    def request(self, req_body=None, method=None,
                header=constants.CONTENT_TYPE_URLENCODE):
        try:
            resp_body = self._request(req_body, method, header)
        except exception.NotAuthorized:
            LOG.debug("Login again because client certification "
                      "may be expired.")
            self._do_setup()
            resp_body = self._request(req_body, method, header)

        return resp_body


class SSHConnector(object):
    def __init__(self, configuration, debug=True):
        super(SSHConnector, self).__init__()
        self.storage_ip = configuration.emc_nas_server
        self.username = configuration.emc_nas_login
        self.password = configuration.emc_nas_password
        self.debug = debug

        self.sshpool = utils.SSHPool(ip=self.storage_ip,
                                     port=22,
                                     conn_timeout=None,
                                     login=self.username,
                                     password=self.password)

    def run_ssh(self, cmd_list, check_exit_code=False):
        command = ' '.join(pipes.quote(cmd_arg) for cmd_arg in cmd_list)

        with self.sshpool.item() as ssh:
            try:
                out, err = processutils.ssh_execute(
                    ssh, command, check_exit_code=check_exit_code)
                self.log_request(command, out, err)

                return out, err
            except processutils.ProcessExecutionError as e:
                with excutils.save_and_reraise_exception():
                    LOG.error('Error running SSH command: %(cmd)s. '
                              'Error: %(excmsg)s.',
                              {'cmd': command, 'excmsg': e})

    def log_request(self, cmd, out, err):
        if not self.debug:
            return

        LOG.debug("\nSSH command: %s.\n", cmd)
        LOG.debug("SSH command output: out=%(out)s, err=%(err)s.\n",
                  {'out': out, 'err': err})



# Copyright (c) 2014 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os
import pipes

from oslo_concurrency import processutils
from oslo_log import log

from manila import exception
from manila.i18n import _
from manila import utils

LOG = log.getLogger(__name__)


def patch(base, *overlays):
    """Recursive dictionary patching."""
    for ovl in overlays:
        for k, v in ovl.items():
            if isinstance(v, dict) and isinstance(base.get(k), dict):
                patch(base[k], v)
            else:
                base[k] = v
    return base


def walk(dct):
    """Recursive iteration over dictionary."""
    for k, v in dct.items():
        if isinstance(v, dict):
            for w in walk(v):
                yield w
        else:
            yield k, v


class RootExecutor(object):
    """Execute wrapper defaulting to root execution."""

    def __init__(self, execute=utils.execute):
        self.execute = execute

    def __call__(self, *args, **kwargs):
        exkwargs = {"run_as_root": True}
        exkwargs.update(kwargs)
        return self.execute(*args, **exkwargs)


class SSHExecutor(object):
    """Callable encapsulating exec through ssh."""

    def __init__(self, *args, **kwargs):
        self.pool = utils.SSHPool(*args, **kwargs)

    def __call__(self, *args, **kwargs):
        # argument with identifier 'run_as_root=' is not accepted by
        # processutils's ssh_execute() method unlike processutils's execute()
        # method. So implement workaround to enable or disable 'run as root'
        # behavior.
        run_as_root = kwargs.pop('run_as_root', False)
        cmd = ' '.join(pipes.quote(a) for a in args)
        if run_as_root:
            cmd = ' '.join(['sudo', cmd])
        ssh = self.pool.get()
        try:
            ret = processutils.ssh_execute(ssh, cmd, **kwargs)
        finally:
            self.pool.put(ssh)
        return ret


def path_from(fpath, *rpath):
    """Return the join of the dir of fpath and rpath in absolute form."""
    return os.path.join(os.path.abspath(os.path.dirname(fpath)), *rpath)


def validate_access_rule(supported_access_types, supported_access_levels,
                         access_rule, abort=False):

    """Validate an access rule.

    :param access_rule: Access rules to be validated.
    :param supported_access_types: List of access types that are regarded
           valid.
    :param supported_access_levels: List of access levels that are
           regarded valid.
    :param abort: a boolean value that indicates if an exception should
                  be raised whether the rule is invalid.
    :return: Boolean.
    """

    errmsg = _("Unsupported access rule of 'type' %(access_type)s, "
               "'level' %(access_level)s, 'to' %(access_to)s: "
               "%(field)s should be one of %(supported)s.")
    access_param = access_rule.to_dict()

    def validate(field, supported_tokens, excinfo):
        if access_rule['access_%s' % field] in supported_tokens:
            return True

        access_param['field'] = field
        access_param['supported'] = ', '.join(
            "'%s'" % x for x in supported_tokens)
        if abort:
            LOG.error(errmsg, access_param)
            raise excinfo['type'](
                **{excinfo['about']: excinfo['details'] % access_param})
        else:
            LOG.warning(errmsg, access_param)
            return False

    valid = True
    valid &= validate(
        'type', supported_access_types,
        {'type': exception.InvalidShareAccess, 'about': "reason",
         'details': _(
             "%(access_type)s; only %(supported)s access type is allowed")})
    valid &= validate(
        'level', supported_access_levels,
        {'type': exception.InvalidShareAccessLevel, 'about': "level",
         'details': "%(access_level)s"})

    return valid


def fixup_access_rule(access_rule):
    """Adjust access rule as required for ganesha to handle it properly.

    :param access_rule: Access rules to be fixed up.
    :return: access_rule
    """
    if access_rule['access_to'] == '0.0.0.0/0':
        access_rule['access_to'] = '0.0.0.0'
        LOG.debug("Set access_to field to '0.0.0.0' in ganesha back end.")

    return access_rule



# Copyright 2013 OpenStack Foundation
# Copyright 2014 Mirantis Inc.
# All Rights Reserved
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock
from neutronclient.common import exceptions as neutron_client_exc
from neutronclient.v2_0 import client as clientv20
from oslo_config import cfg

from manila.db import base
from manila import exception
from manila.network.neutron import api as neutron_api
from manila.network.neutron import constants as neutron_constants
from manila import test
from manila.tests.db import fakes
from manila.tests import utils as test_utils

CONF = cfg.CONF


class FakeNeutronClient(object):

    def create_port(self, body):
        return body

    def delete_port(self, port_id):
        pass

    def show_port(self, port_id):
        pass

    def list_ports(self, **search_opts):
        pass

    def list_networks(self):
        pass

    def show_network(self, network_uuid):
        pass

    def show_subnet(self, subnet_uuid):
        pass

    def create_router(self, body):
        return body

    def list_routers(self):
        pass

    def create_network(self, body):
        return body

    def create_subnet(self, body):
        return body

    def update_port(self, port_id, body):
        return body

    def add_interface_router(self, router_id, subnet_id, port_id):
        pass

    def update_router(self, router_id, body):
        return body

    def show_router(self, router_id):
        pass

    def list_extensions(self):
        pass


class NeutronclientTestCase(test.TestCase):

    def test_no_auth_obj(self):
        mock_client_loader = self.mock_object(
            neutron_api.client_auth, 'AuthClientLoader')
        fake_context = 'fake_context'
        data = {
            'neutron': {
                'endpoint_type': 'foo_endpoint_type',
                'region_name': 'foo_region_name',
            }
        }

        self.client = None
        with test_utils.create_temp_config_with_opts(data):
            self.client = neutron_api.API()
            self.client.get_client(fake_context)

        mock_client_loader.assert_called_once_with(
            client_class=neutron_api.clientv20.Client,
            exception_module=neutron_api.neutron_client_exc,
            cfg_group=neutron_api.NEUTRON_GROUP
        )
        mock_client_loader.return_value.get_client.assert_called_once_with(
            self.client,
            fake_context,
            endpoint_type=data['neutron']['endpoint_type'],
            region_name=data['neutron']['region_name'],
        )

    def test_with_auth_obj(self):
        fake_context = 'fake_context'
        data = {
            'neutron': {
                'endpoint_type': 'foo_endpoint_type',
                'region_name': 'foo_region_name',
            }
        }

        self.client = None
        with test_utils.create_temp_config_with_opts(data):
            self.client = neutron_api.API()
            self.client.auth_obj = type(
                'FakeAuthObj', (object, ), {'get_client': mock.Mock()})
            self.client.get_client(fake_context)

        self.client.auth_obj.get_client.assert_called_once_with(
            self.client,
            fake_context,
            endpoint_type=data['neutron']['endpoint_type'],
            region_name=data['neutron']['region_name'],
        )


class NeutronApiTest(test.TestCase):

    def setUp(self):
        super(NeutronApiTest, self).setUp()
        self.mock_object(base, 'Base', fakes.FakeModel)
        self.mock_object(
            clientv20, 'Client', mock.Mock(return_value=FakeNeutronClient()))
        self.neutron_api = neutron_api.API()

    def test_create_api_object(self):
        # instantiate Neutron API object
        neutron_api_instance = neutron_api.API()

        # Verify results
        self.assertTrue(hasattr(neutron_api_instance, 'client'))
        self.assertTrue(hasattr(neutron_api_instance, 'configuration'))
        self.assertEqual('DEFAULT', neutron_api_instance.config_group_name)

    def test_create_port_with_all_args(self):
        # Set up test data
        self.mock_object(self.neutron_api, '_has_port_binding_extension',
                         mock.Mock(return_value=True))
        port_args = {
            'tenant_id': 'test tenant', 'network_id': 'test net',
            'host_id': 'test host', 'subnet_id': 'test subnet',
            'fixed_ip': 'test ip', 'device_owner': 'test owner',
            'device_id': 'test device', 'mac_address': 'test mac',
            'security_group_ids': 'test group',
            'dhcp_opts': 'test dhcp',
        }

        # Execute method 'create_port'
        port = self.neutron_api.create_port(**port_args)

        # Verify results
        self.assertEqual(port_args['tenant_id'], port['tenant_id'])
        self.assertEqual(port_args['network_id'], port['network_id'])
        self.assertEqual(port_args['host_id'], port['binding:host_id'])
        self.assertEqual(port_args['subnet_id'],
                         port['fixed_ips'][0]['subnet_id'])
        self.assertEqual(port_args['fixed_ip'],
                         port['fixed_ips'][0]['ip_address'])
        self.assertEqual(port_args['device_owner'], port['device_owner'])
        self.assertEqual(port_args['device_id'], port['device_id'])
        self.assertEqual(port_args['mac_address'], port['mac_address'])
        self.assertEqual(port_args['security_group_ids'],
                         port['security_groups'])
        self.assertEqual(port_args['dhcp_opts'], port['extra_dhcp_opts'])
        self.neutron_api._has_port_binding_extension.assert_called_once_with()
        self.assertTrue(clientv20.Client.called)

    def test_create_port_with_required_args(self):
        # Set up test data
        port_args = {'tenant_id': 'test tenant', 'network_id': 'test net'}

        # Execute method 'create_port'
        port = self.neutron_api.create_port(**port_args)

        # Verify results
        self.assertEqual(port_args['tenant_id'], port['tenant_id'])
        self.assertEqual(port_args['network_id'],
                         port['network_id'])
        self.assertTrue(clientv20.Client.called)

    def test_create_port_with_additional_kwargs(self):
        # Set up test data
        port_args = {'tenant_id': 'test tenant', 'network_id': 'test net',
                     'binding_arg': 'foo'}

        # Execute method 'create_port'
        port = self.neutron_api.create_port(**port_args)

        # Verify results
        self.assertEqual(port_args['tenant_id'], port['tenant_id'])
        self.assertEqual(port_args['network_id'],
                         port['network_id'])
        self.assertEqual(port_args['binding_arg'],
                         port['binding_arg'])
        self.assertTrue(clientv20.Client.called)

    def test_create_port_with_host_id_no_binding_ext(self):
        self.mock_object(self.neutron_api, '_has_port_binding_extension',
                         mock.Mock(return_value=False))
        port_args = {
            'tenant_id': 'test tenant',
            'network_id': 'test net',
            'host_id': 'foohost'
        }

        self.assertRaises(exception.NetworkException,
                          self.neutron_api.create_port, **port_args)

    @mock.patch.object(neutron_api.LOG, 'exception', mock.Mock())
    def test_create_port_exception(self):
        self.mock_object(
            self.neutron_api.client, 'create_port',
            mock.Mock(side_effect=neutron_client_exc.NeutronClientException))
        port_args = {'tenant_id': 'test tenant', 'network_id': 'test net'}

        # Execute method 'create_port'
        self.assertRaises(exception.NetworkException,
                          self.neutron_api.create_port,
                          **port_args)

        # Verify results
        self.assertTrue(neutron_api.LOG.exception.called)
        self.assertTrue(clientv20.Client.called)
        self.assertTrue(self.neutron_api.client.create_port.called)

    @mock.patch.object(neutron_api.LOG, 'exception', mock.Mock())
    def test_create_port_exception_status_409(self):
        # Set up test data
        self.mock_object(
            self.neutron_api.client, 'create_port',
            mock.Mock(side_effect=neutron_client_exc.NeutronClientException(
                status_code=409)))
        port_args = {'tenant_id': 'test tenant', 'network_id': 'test net'}

        # Execute method 'create_port'
        self.assertRaises(exception.PortLimitExceeded,
                          self.neutron_api.create_port,
                          **port_args)

        # Verify results
        self.assertTrue(neutron_api.LOG.exception.called)
        self.assertTrue(clientv20.Client.called)
        self.assertTrue(self.neutron_api.client.create_port.called)

    def test_delete_port(self):
        # Set up test data
        self.mock_object(self.neutron_api.client, 'delete_port')
        port_id = 'test port id'

        # Execute method 'delete_port'
        self.neutron_api.delete_port(port_id)

        # Verify results
        self.neutron_api.client.delete_port.assert_called_once_with(port_id)
        self.assertTrue(clientv20.Client.called)

    def test_list_ports(self):
        # Set up test data
        search_opts = {'test_option': 'test_value'}
        fake_ports = [{'fake port': 'fake port info'}]
        self.mock_object(
            self.neutron_api.client, 'list_ports',
            mock.Mock(return_value={'ports': fake_ports}))

        # Execute method 'list_ports'
        ports = self.neutron_api.list_ports(**search_opts)

        # Verify results
        self.assertEqual(fake_ports, ports)
        self.assertTrue(clientv20.Client.called)
        self.neutron_api.client.list_ports.assert_called_once_with(
            **search_opts)

    def test_show_port(self):
        # Set up test data
        port_id = 'test port id'
        fake_port = {'fake port': 'fake port info'}
        self.mock_object(
            self.neutron_api.client, 'show_port',
            mock.Mock(return_value={'port': fake_port}))

        # Execute method 'show_port'
        port = self.neutron_api.show_port(port_id)

        # Verify results
        self.assertEqual(fake_port, port)
        self.assertTrue(clientv20.Client.called)
        self.neutron_api.client.show_port.assert_called_once_with(port_id)

    def test_get_network(self):
        # Set up test data
        network_id = 'test network id'
        fake_network = {'fake network': 'fake network info'}
        self.mock_object(
            self.neutron_api.client, 'show_network',
            mock.Mock(return_value={'network': fake_network}))

        # Execute method 'get_network'
        network = self.neutron_api.get_network(network_id)

        # Verify results
        self.assertEqual(fake_network, network)
        self.assertTrue(clientv20.Client.called)
        self.neutron_api.client.show_network.assert_called_once_with(
            network_id)

    def test_get_subnet(self):
        # Set up test data
        subnet_id = 'fake subnet id'
        self.mock_object(
            self.neutron_api.client, 'show_subnet',
            mock.Mock(return_value={'subnet': {}}))

        # Execute method 'get_subnet'
        subnet = self.neutron_api.get_subnet(subnet_id)

        # Verify results
        self.assertEqual({}, subnet)
        self.assertTrue(clientv20.Client.called)
        self.neutron_api.client.show_subnet.assert_called_once_with(
            subnet_id)

    def test_get_all_network(self):
        # Set up test data
        fake_networks = [{'fake network': 'fake network info'}]
        self.mock_object(
            self.neutron_api.client, 'list_networks',
            mock.Mock(return_value={'networks': fake_networks}))

        # Execute method 'get_all_networks'
        networks = self.neutron_api.get_all_networks()

        # Verify results
        self.assertEqual(fake_networks, networks)
        self.assertTrue(clientv20.Client.called)
        self.neutron_api.client.list_networks.assert_called_once_with()

    def test_list_extensions(self):
        # Set up test data
        extensions = [
            {'name': neutron_constants.PORTBINDING_EXT},
            {'name': neutron_constants.PROVIDER_NW_EXT},
        ]
        self.mock_object(
            self.neutron_api.client, 'list_extensions',
            mock.Mock(return_value={'extensions': extensions}))

        # Execute method 'list_extensions'
        result = self.neutron_api.list_extensions()

        # Verify results
        self.assertTrue(clientv20.Client.called)
        self.neutron_api.client.list_extensions.assert_called_once_with()
        self.assertIn(neutron_constants.PORTBINDING_EXT, result)
        self.assertIn(neutron_constants.PROVIDER_NW_EXT, result)
        self.assertEqual(
            extensions[0], result[neutron_constants.PORTBINDING_EXT])
        self.assertEqual(
            extensions[1], result[neutron_constants.PROVIDER_NW_EXT])

    def test_create_network(self):
        # Set up test data
        net_args = {'tenant_id': 'test tenant', 'name': 'test name'}

        # Execute method 'network_create'
        network = self.neutron_api.network_create(**net_args)

        # Verify results
        self.assertEqual(net_args['tenant_id'], network['tenant_id'])
        self.assertEqual(net_args['name'], network['name'])
        self.assertTrue(clientv20.Client.called)

    def test_create_subnet(self):
        # Set up test data
        subnet_args = {
            'tenant_id': 'test tenant',
            'name': 'test name',
            'net_id': 'test net id',
            'cidr': '10.0.0.0/24',
        }

        # Execute method 'subnet_create'
        subnet = self.neutron_api.subnet_create(**subnet_args)

        # Verify results
        self.assertEqual(subnet_args['tenant_id'], subnet['tenant_id'])
        self.assertEqual(subnet_args['name'], subnet['name'])
        self.assertTrue(clientv20.Client.called)

    def test_create_router(self):
        # Set up test data
        router_args = {'tenant_id': 'test tenant', 'name': 'test name'}

        # Execute method 'router_create'
        router = self.neutron_api.router_create(**router_args)

        # Verify results
        self.assertEqual(router_args['tenant_id'], router['tenant_id'])
        self.assertEqual(router_args['name'], router['name'])
        self.assertTrue(clientv20.Client.called)

    def test_list_routers(self):
        # Set up test data
        fake_routers = [{'fake router': 'fake router info'}]
        self.mock_object(
            self.neutron_api.client, 'list_routers',
            mock.Mock(return_value={'routers': fake_routers}))

        # Execute method 'router_list'
        networks = self.neutron_api.router_list()

        # Verify results
        self.assertEqual(fake_routers, networks)
        self.assertTrue(clientv20.Client.called)
        self.neutron_api.client.list_routers.assert_called_once_with()

    def test_create_network_exception(self):
        # Set up test data
        net_args = {'tenant_id': 'test tenant', 'name': 'test name'}
        self.mock_object(
            self.neutron_api.client, 'create_network',
            mock.Mock(side_effect=neutron_client_exc.NeutronClientException))

        # Execute method 'network_create'
        self.assertRaises(
            exception.NetworkException,
            self.neutron_api.network_create,
            **net_args)

        # Verify results
        self.neutron_api.client.create_network.assert_called_once_with(
            {'network': net_args})
        self.assertTrue(clientv20.Client.called)

    def test_create_subnet_exception(self):
        # Set up test data
        subnet_args = {
            'tenant_id': 'test tenant',
            'name': 'test name',
            'net_id': 'test net id',
            'cidr': '10.0.0.0/24',
        }
        self.mock_object(
            self.neutron_api.client, 'create_subnet',
            mock.Mock(side_effect=neutron_client_exc.NeutronClientException))

        # Execute method 'subnet_create'
        self.assertRaises(
            exception.NetworkException,
            self.neutron_api.subnet_create,
            **subnet_args)

        # Verify results
        expected_data = {
            'network_id': subnet_args['net_id'],
            'tenant_id': subnet_args['tenant_id'],
            'cidr': subnet_args['cidr'],
            'name': subnet_args['name'],
            'ip_version': 4,
        }
        self.neutron_api.client.create_subnet.assert_called_once_with(
            {'subnet': expected_data})
        self.assertTrue(clientv20.Client.called)

    def test_create_router_exception(self):
        # Set up test data
        router_args = {'tenant_id': 'test tenant', 'name': 'test name'}
        self.mock_object(
            self.neutron_api.client, 'create_router',
            mock.Mock(side_effect=neutron_client_exc.NeutronClientException))

        # Execute method 'router_create'
        self.assertRaises(
            exception.NetworkException,
            self.neutron_api.router_create,
            **router_args)

        # Verify results
        self.neutron_api.client.create_router.assert_called_once_with(
            {'router': router_args})
        self.assertTrue(clientv20.Client.called)

    def test_update_port_fixed_ips(self):
        # Set up test data
        port_id = 'test_port'
        fixed_ips = {'fixed_ips': [{'subnet_id': 'test subnet'}]}

        # Execute method 'update_port_fixed_ips'
        port = self.neutron_api.update_port_fixed_ips(port_id, fixed_ips)

        # Verify results
        self.assertEqual(fixed_ips, port)
        self.assertTrue(clientv20.Client.called)

    def test_update_port_fixed_ips_exception(self):
        # Set up test data
        port_id = 'test_port'
        fixed_ips = {'fixed_ips': [{'subnet_id': 'test subnet'}]}
        self.mock_object(
            self.neutron_api.client, 'update_port',
            mock.Mock(side_effect=neutron_client_exc.NeutronClientException))

        # Execute method 'update_port_fixed_ips'
        self.assertRaises(
            exception.NetworkException,
            self.neutron_api.update_port_fixed_ips,
            port_id, fixed_ips)

        # Verify results
        self.neutron_api.client.update_port.assert_called_once_with(
            port_id, {'port': fixed_ips})
        self.assertTrue(clientv20.Client.called)

    def test_router_update_routes(self):
        # Set up test data
        router_id = 'test_router'
        routes = {
            'routes': [
                {'destination': '0.0.0.0/0', 'nexthop': '8.8.8.8', },
            ],
        }

        # Execute method 'router_update_routes'
        router = self.neutron_api.router_update_routes(router_id, routes)

        # Verify results
        self.assertEqual(routes, router)
        self.assertTrue(clientv20.Client.called)

    def test_router_update_routes_exception(self):
        # Set up test data
        router_id = 'test_router'
        routes = {
            'routes': [
                {'destination': '0.0.0.0/0', 'nexthop': '8.8.8.8', },
            ],
        }
        self.mock_object(
            self.neutron_api.client, 'update_router',
            mock.Mock(side_effect=neutron_client_exc.NeutronClientException))

        # Execute method 'router_update_routes'
        self.assertRaises(
            exception.NetworkException,
            self.neutron_api.router_update_routes,
            router_id, routes)

        # Verify results
        self.neutron_api.client.update_router.assert_called_once_with(
            router_id, {'router': routes})
        self.assertTrue(clientv20.Client.called)

    def test_show_router(self):
        # Set up test data
        router_id = 'test router id'
        fake_router = {'fake router': 'fake router info'}
        self.mock_object(
            self.neutron_api.client, 'show_router',
            mock.Mock(return_value={'router': fake_router}))

        # Execute method 'show_router'
        port = self.neutron_api.show_router(router_id)

        # Verify results
        self.assertEqual(fake_router, port)
        self.assertTrue(clientv20.Client.called)
        self.neutron_api.client.show_router.assert_called_once_with(router_id)

    def test_router_add_interface(self):
        # Set up test data
        router_id = 'test port id'
        subnet_id = 'test subnet id'
        port_id = 'test port id'
        self.mock_object(self.neutron_api.client, 'add_interface_router')

        # Execute method 'router_add_interface'
        self.neutron_api.router_add_interface(router_id, subnet_id, port_id)

        # Verify results
        self.neutron_api.client.add_interface_router.assert_called_once_with(
            port_id, {'subnet_id': subnet_id, 'port_id': port_id})
        self.assertTrue(clientv20.Client.called)

    def test_router_add_interface_exception(self):
        # Set up test data
        router_id = 'test port id'
        subnet_id = 'test subnet id'
        port_id = 'test port id'
        self.mock_object(
            self.neutron_api.client, 'add_interface_router',
            mock.Mock(side_effect=neutron_client_exc.NeutronClientException))

        # Execute method 'router_add_interface'
        self.assertRaises(
            exception.NetworkException,
            self.neutron_api.router_add_interface,
            router_id, subnet_id, port_id)

        # Verify results
        self.neutron_api.client.add_interface_router.assert_called_once_with(
            router_id, {'subnet_id': subnet_id, 'port_id': port_id})
        self.assertTrue(clientv20.Client.called)

    def test_admin_project_id_exist(self):
        fake_admin_project_id = 'fake_admin_project_id_value'
        self.neutron_api.client.httpclient = mock.Mock()
        self.neutron_api.client.httpclient.auth_token = mock.Mock()
        self.neutron_api.client.httpclient.get_project_id = mock.Mock(
            return_value=fake_admin_project_id)

        admin_project_id = self.neutron_api.admin_project_id

        self.assertEqual(fake_admin_project_id, admin_project_id)
        self.neutron_api.client.httpclient.auth_token.called

    def test_admin_project_id_not_exist(self):
        fake_admin_project_id = 'fake_admin_project_id_value'
        self.neutron_api.client.httpclient = mock.Mock()
        self.neutron_api.client.httpclient.auth_token = mock.Mock(
            return_value=None)
        self.neutron_api.client.httpclient.authenticate = mock.Mock()
        self.neutron_api.client.httpclient.get_project_id = mock.Mock(
            return_value=fake_admin_project_id)

        admin_project_id = self.neutron_api.admin_project_id

        self.assertEqual(fake_admin_project_id, admin_project_id)
        self.neutron_api.client.httpclient.auth_token.called
        self.neutron_api.client.httpclient.authenticate.called

    def test_admin_project_id_not_exist_with_failure(self):
        self.neutron_api.client.httpclient = mock.Mock()
        self.neutron_api.client.httpclient.auth_token = None
        self.neutron_api.client.httpclient.authenticate = mock.Mock(
            side_effect=neutron_client_exc.NeutronClientException)
        self.neutron_api.client.httpclient.auth_tenant_id = mock.Mock()

        try:
            self.neutron_api.admin_project_id
        except exception.NetworkException:
            pass
        else:
            raise Exception('Expected error was not raised')

        self.assertTrue(self.neutron_api.client.httpclient.authenticate.called)
        self.assertFalse(
            self.neutron_api.client.httpclient.auth_tenant_id.called)

    def test_get_all_admin_project_networks(self):
        fake_networks = {'networks': ['fake_net_1', 'fake_net_2']}
        self.mock_object(
            self.neutron_api.client, 'list_networks',
            mock.Mock(return_value=fake_networks))
        self.neutron_api.client.httpclient = mock.Mock()
        self.neutron_api.client.httpclient.auth_token = mock.Mock()
        self.neutron_api.client.httpclient.auth_tenant_id = mock.Mock()

        networks = self.neutron_api.get_all_admin_project_networks()

        self.assertEqual(fake_networks['networks'], networks)
        self.neutron_api.client.httpclient.auth_token.called
        self.neutron_api.client.httpclient.auth_tenant_id.called
        self.neutron_api.client.list_networks.assert_called_once_with(
            tenant_id=self.neutron_api.admin_project_id, shared=False)



# Copyright 2015, Hitachi Data Systems.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Client side of the data manager RPC API.
"""

from oslo_config import cfg
import oslo_messaging as messaging

from manila import rpc

CONF = cfg.CONF


class DataAPI(object):
    """Client side of the data RPC API.

    API version history:

        1.0 - Initial version,
              Add migration_start(),
              data_copy_cancel(),
              data_copy_get_progress()
    """

    BASE_RPC_API_VERSION = '1.0'

    def __init__(self):
        super(DataAPI, self).__init__()
        target = messaging.Target(topic=CONF.data_topic,
                                  version=self.BASE_RPC_API_VERSION)
        self.client = rpc.get_client(target, version_cap='1.0')

    def migration_start(self, context, share_id, ignore_list,
                        share_instance_id, dest_share_instance_id,
                        connection_info_src, connection_info_dest):
        call_context = self.client.prepare(version='1.0')
        call_context.cast(
            context,
            'migration_start',
            share_id=share_id,
            ignore_list=ignore_list,
            share_instance_id=share_instance_id,
            dest_share_instance_id=dest_share_instance_id,
            connection_info_src=connection_info_src,
            connection_info_dest=connection_info_dest)

    def data_copy_cancel(self, context, share_id):
        call_context = self.client.prepare(version='1.0')
        call_context.call(context, 'data_copy_cancel', share_id=share_id)

    def data_copy_get_progress(self, context, share_id):
        call_context = self.client.prepare(version='1.0')
        return call_context.call(context, 'data_copy_get_progress',
                                 share_id=share_id)



# Copyright 2011-2012 OpenStack LLC.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""
Tests For Capacity Weigher.
"""

import ddt
import mock
from oslo_config import cfg

from manila import context
from manila.scheduler.weighers import base_host
from manila.scheduler.weighers import capacity
from manila.share import utils
from manila import test
from manila.tests.scheduler import fakes

CONF = cfg.CONF


@ddt.ddt
class CapacityWeigherTestCase(test.TestCase):
    def setUp(self):
        super(CapacityWeigherTestCase, self).setUp()
        self.host_manager = fakes.FakeHostManager()
        self.weight_handler = base_host.HostWeightHandler(
            'manila.scheduler.weighers')

    def _get_weighed_host(self, hosts, weight_properties=None, index=0):
        if weight_properties is None:
            weight_properties = {'size': 1}
        return self.weight_handler.get_weighed_objects(
            [capacity.CapacityWeigher],
            hosts,
            weight_properties)[index]

    @mock.patch('manila.db.api.IMPL.service_get_all_by_topic')
    def _get_all_hosts(self, _mock_service_get_all_by_topic, disabled=False):
        ctxt = context.get_admin_context()
        fakes.mock_host_manager_db_calls(_mock_service_get_all_by_topic,
                                         disabled=disabled)
        host_states = self.host_manager.get_all_host_states_share(ctxt)
        _mock_service_get_all_by_topic.assert_called_once_with(
            ctxt, CONF.share_topic)
        return host_states

    # NOTE(xyang): If thin_provisioning = True and
    # max_over_subscription_ratio >= 1, use the following formula:
    # free = math.floor(total * host_state.max_over_subscription_ratio
    #        - host_state.provisioned_capacity_gb
    #        - total * reserved)
    # Otherwise, use the following formula:
    # free = math.floor(free_space - total * reserved)

    @ddt.data(
        {'cap_thin': '<is> True',
         'cap_thin_key': 'capabilities:thin_provisioning',
         'winner': 'host2'},
        {'cap_thin': '<is> False',
         'cap_thin_key': 'thin_provisioning',
         'winner': 'host1'},
        {'cap_thin': 'True',
         'cap_thin_key': 'capabilities:thin_provisioning',
         'winner': 'host2'},
        {'cap_thin': 'False',
         'cap_thin_key': 'thin_provisioning',
         'winner': 'host1'},
        {'cap_thin': 'true',
         'cap_thin_key': 'capabilities:thin_provisioning',
         'winner': 'host2'},
        {'cap_thin': 'false',
         'cap_thin_key': 'thin_provisioning',
         'winner': 'host1'},
        {'cap_thin': None,
         'cap_thin_key': None,
         'winner': 'host2'},
    )
    @ddt.unpack
    def test_default_of_spreading_first(self, cap_thin, cap_thin_key,
                                        winner):
        hosts = self._get_all_hosts()  # pylint: disable=no-value-for-parameter

        # Results for the 1st test
        # {'capabilities:thin_provisioning': '<is> True'}:
        # host1: thin_provisioning = False
        #        free_capacity_gb = 1024
        #        free = math.floor(1024 - 1024 * 0.1) = 921.0
        #        weight = 0.40
        # host2: thin_provisioning = True
        #        max_over_subscription_ratio = 2.0
        #        free_capacity_gb = 300
        #        free = math.floor(2048 * 2.0 - 1748 - 2048 * 0.1)=2143.0
        #        weight = 1.0
        # host3: thin_provisioning = [False]
        #        free_capacity_gb = 512
        #        free = math.floor(256 - 512 * 0)=256.0
        #        weight = 0.08
        # host4: thin_provisioning = [True]
        #        max_over_subscription_ratio = 1.0
        #        free_capacity_gb = 200
        #        free = math.floor(2048 * 1.0 - 1848 - 2048 * 0.05) = 97.0
        #        weight = 0.0
        # host5: thin_provisioning = [True, False]
        #        max_over_subscription_ratio = 1.5
        #        free_capacity_gb = 500
        #        free = math.floor(2048 * 1.5 - 1548 - 2048 * 0.05) = 1421.0
        #        weight = 0.65
        # host6: thin_provisioning = False
        #        free = inf
        #        weight = 0.0

        # so, host2 should win:
        weight_properties = {
            'size': 1,
            'share_type': {
                'extra_specs': {
                    cap_thin_key: cap_thin,
                }
            }
        }
        weighed_host = self._get_weighed_host(
            hosts,
            weight_properties=weight_properties)
        self.assertEqual(1.0, weighed_host.weight)
        self.assertEqual(
            winner, utils.extract_host(weighed_host.obj.host))

    def test_unknown_is_last(self):
        hosts = self._get_all_hosts()  # pylint: disable=no-value-for-parameter

        last_host = self._get_weighed_host(hosts, index=-1)
        self.assertEqual(
            'host6', utils.extract_host(last_host.obj.host))
        self.assertEqual(0.0, last_host.weight)

    @ddt.data(
        {'cap_thin': '<is> True',
         'cap_thin_key': 'capabilities:thin_provisioning',
         'winner': 'host4'},
        {'cap_thin': '<is> False',
         'cap_thin_key': 'thin_provisioning',
         'winner': 'host2'},
        {'cap_thin': 'True',
         'cap_thin_key': 'capabilities:thin_provisioning',
         'winner': 'host4'},
        {'cap_thin': 'False',
         'cap_thin_key': 'thin_provisioning',
         'winner': 'host2'},
        {'cap_thin': 'true',
         'cap_thin_key': 'capabilities:thin_provisioning',
         'winner': 'host4'},
        {'cap_thin': 'false',
         'cap_thin_key': 'thin_provisioning',
         'winner': 'host2'},
        {'cap_thin': None,
         'cap_thin_key': None,
         'winner': 'host4'},
    )
    @ddt.unpack
    def test_capacity_weight_multiplier_negative_1(self, cap_thin,
                                                   cap_thin_key,
                                                   winner):
        self.flags(capacity_weight_multiplier=-1.0)
        hosts = self._get_all_hosts()  # pylint: disable=no-value-for-parameter

        # Results for the 1st test
        # {'capabilities:thin_provisioning': '<is> True'}:
        # host1: thin_provisioning = False
        #        free_capacity_gb = 1024
        #        free = math.floor(1024 - 1024 * 0.1) = 921.0
        #        free * (-1) = -921.0
        #        weight = -0.40
        # host2: thin_provisioning = True
        #        max_over_subscription_ratio = 2.0
        #        free_capacity_gb = 300
        #        free = math.floor(2048 * 2.0-1748-2048 * 0.1) = 2143.0
        #        free * (-1) = -2143.0
        #        weight = -1.0
        # host3: thin_provisioning = [False]
        #        free_capacity_gb = 512
        #        free = math.floor(256 - 512 * 0) = 256.0
        #        free * (-1) = -256.0
        #        weight = -0.08
        # host4: thin_provisioning = [True]
        #        max_over_subscription_ratio = 1.0
        #        free_capacity_gb = 200
        #        free = math.floor(2048 * 1.0 - 1848 - 2048 * 0.05) = 97.0
        #        free * (-1) = -97.0
        #        weight = 0.0
        # host5: thin_provisioning = [True, False]
        #        max_over_subscription_ratio = 1.5
        #        free_capacity_gb = 500
        #        free = math.floor(2048 * 1.5 - 1548 - 2048 * 0.05) = 1421.0
        #        free * (-1) = -1421.0
        #        weight = -0.65
        # host6: thin_provisioning = False
        #        free = inf
        #        free * (-1) = -inf
        #        weight = 0.0

        # so, host4 should win:
        weight_properties = {
            'size': 1,
            'share_type': {
                'extra_specs': {
                    cap_thin_key: cap_thin,
                }
            }
        }
        weighed_host = self._get_weighed_host(
            hosts,
            weight_properties=weight_properties)
        self.assertEqual(0.0, weighed_host.weight)
        self.assertEqual(
            winner, utils.extract_host(weighed_host.obj.host))

    @ddt.data(
        {'cap_thin': '<is> True',
         'cap_thin_key': 'capabilities:thin_provisioning',
         'winner': 'host2'},
        {'cap_thin': '<is> False',
         'cap_thin_key': 'thin_provisioning',
         'winner': 'host1'},
        {'cap_thin': 'True',
         'cap_thin_key': 'capabilities:thin_provisioning',
         'winner': 'host2'},
        {'cap_thin': 'False',
         'cap_thin_key': 'thin_provisioning',
         'winner': 'host1'},
        {'cap_thin': 'true',
         'cap_thin_key': 'capabilities:thin_provisioning',
         'winner': 'host2'},
        {'cap_thin': 'false',
         'cap_thin_key': 'thin_provisioning',
         'winner': 'host1'},
        {'cap_thin': None,
         'cap_thin_key': None,
         'winner': 'host2'},
    )
    @ddt.unpack
    def test_capacity_weight_multiplier_2(self, cap_thin, cap_thin_key,
                                          winner):
        self.flags(capacity_weight_multiplier=2.0)
        hosts = self._get_all_hosts()  # pylint: disable=no-value-for-parameter

        # Results for the 1st test
        # {'capabilities:thin_provisioning': '<is> True'}:
        # host1: thin_provisioning = False
        #        free_capacity_gb = 1024
        #        free = math.floor(1024-1024*0.1) = 921.0
        #        free * 2 = 1842.0
        #        weight = 0.81
        # host2: thin_provisioning = True
        #        max_over_subscription_ratio = 2.0
        #        free_capacity_gb = 300
        #        free = math.floor(2048 * 2.0 - 1748 - 2048 * 0.1) = 2143.0
        #        free * 2 = 4286.0
        #        weight = 2.0
        # host3: thin_provisioning = [False]
        #        free_capacity_gb = 512
        #        free = math.floor(256 - 512 * 0) = 256.0
        #        free * 2 = 512.0
        #        weight = 0.16
        # host4: thin_provisioning = [True]
        #        max_over_subscription_ratio = 1.0
        #        free_capacity_gb = 200
        #        free = math.floor(2048 * 1.0 - 1848 - 2048 * 0.05) = 97.0
        #        free * 2 = 194.0
        #        weight = 0.0
        # host5: thin_provisioning = [True, False]
        #        max_over_subscription_ratio = 1.5
        #        free_capacity_gb = 500
        #        free = math.floor(2048 * 1.5 - 1548 - 2048 * 0.05) = 1421.0
        #        free * 2 = 2842.0
        #        weight = 1.29
        # host6: thin_provisioning = False
        #        free = inf
        #        weight = 0.0

        # so, host2 should win:
        weight_properties = {
            'size': 1,
            'share_type': {
                'extra_specs': {
                    cap_thin_key: cap_thin,
                }
            }
        }
        weighed_host = self._get_weighed_host(
            hosts, weight_properties=weight_properties)
        self.assertEqual(2.0, weighed_host.weight)
        self.assertEqual(
            winner, utils.extract_host(weighed_host.obj.host))



# Copyright (c) 2015 Hitachi Data Systems.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""Testing of SQLAlchemy model classes."""

import ddt

from manila.common import constants
from manila import context
from manila.db.sqlalchemy import api as db_api
from manila import test
from manila.tests import db_utils


@ddt.ddt
class ShareTestCase(test.TestCase):
    """Testing of SQLAlchemy Share model class."""

    @ddt.data(constants.STATUS_MANAGE_ERROR, constants.STATUS_CREATING,
              constants.STATUS_EXTENDING, constants.STATUS_DELETING,
              constants.STATUS_EXTENDING_ERROR,
              constants.STATUS_ERROR_DELETING, constants.STATUS_MANAGING,
              constants.STATUS_MANAGE_ERROR)
    def test_share_instance_available(self, status):

        instance_list = [
            db_utils.create_share_instance(status=constants.STATUS_AVAILABLE,
                                           share_id='fake_id'),
            db_utils.create_share_instance(status=status,
                                           share_id='fake_id')
        ]

        share1 = db_utils.create_share(instances=instance_list)
        share2 = db_utils.create_share(instances=list(reversed(instance_list)))

        self.assertEqual(constants.STATUS_AVAILABLE, share1.instance['status'])
        self.assertEqual(constants.STATUS_AVAILABLE, share2.instance['status'])

    @ddt.data([constants.STATUS_MANAGE_ERROR, constants.STATUS_CREATING],
              [constants.STATUS_ERROR_DELETING, constants.STATUS_DELETING],
              [constants.STATUS_ERROR, constants.STATUS_MANAGING],
              [constants.STATUS_UNMANAGE_ERROR, constants.STATUS_UNMANAGING],
              [constants.STATUS_INACTIVE, constants.STATUS_EXTENDING],
              [constants.STATUS_SHRINKING_ERROR, constants.STATUS_SHRINKING])
    @ddt.unpack
    def test_share_instance_not_transitional(self, status, trans_status):

        instance_list = [
            db_utils.create_share_instance(status=status,
                                           share_id='fake_id'),
            db_utils.create_share_instance(status=trans_status,
                                           share_id='fake_id')
        ]

        share1 = db_utils.create_share(instances=instance_list)
        share2 = db_utils.create_share(instances=list(reversed(instance_list)))

        self.assertEqual(status, share1.instance['status'])
        self.assertEqual(status, share2.instance['status'])

    def test_share_instance_creating(self):

        share = db_utils.create_share(status=constants.STATUS_CREATING)

        self.assertEqual(constants.STATUS_CREATING, share.instance['status'])

    @ddt.data(constants.STATUS_REPLICATION_CHANGE, constants.STATUS_AVAILABLE,
              constants.STATUS_ERROR, constants.STATUS_CREATING)
    def test_share_instance_reverting(self, status):

        instance_list = [
            db_utils.create_share_instance(
                status=constants.STATUS_REVERTING,
                share_id='fake_id'),
            db_utils.create_share_instance(
                status=status, share_id='fake_id'),
            db_utils.create_share_instance(
                status=constants.STATUS_ERROR_DELETING, share_id='fake_id'),
        ]

        share1 = db_utils.create_share(instances=instance_list)
        share2 = db_utils.create_share(instances=list(reversed(instance_list)))

        self.assertEqual(
            constants.STATUS_REVERTING, share1.instance['status'])
        self.assertEqual(
            constants.STATUS_REVERTING, share2.instance['status'])

    @ddt.data(constants.STATUS_AVAILABLE, constants.STATUS_ERROR,
              constants.STATUS_CREATING)
    def test_share_instance_replication_change(self, status):

        instance_list = [
            db_utils.create_share_instance(
                status=constants.STATUS_REPLICATION_CHANGE,
                share_id='fake_id'),
            db_utils.create_share_instance(
                status=status, share_id='fake_id'),
            db_utils.create_share_instance(
                status=constants.STATUS_ERROR_DELETING, share_id='fake_id')
        ]

        share1 = db_utils.create_share(instances=instance_list)
        share2 = db_utils.create_share(instances=list(reversed(instance_list)))

        self.assertEqual(
            constants.STATUS_REPLICATION_CHANGE, share1.instance['status'])
        self.assertEqual(
            constants.STATUS_REPLICATION_CHANGE, share2.instance['status'])

    def test_share_instance_prefer_active_instance(self):

        instance_list = [
            db_utils.create_share_instance(
                status=constants.STATUS_AVAILABLE,
                share_id='fake_id',
                replica_state=constants.REPLICA_STATE_IN_SYNC),
            db_utils.create_share_instance(
                status=constants.STATUS_CREATING,
                share_id='fake_id',
                replica_state=constants.REPLICA_STATE_OUT_OF_SYNC),
            db_utils.create_share_instance(
                status=constants.STATUS_ERROR, share_id='fake_id',
                replica_state=constants.REPLICA_STATE_ACTIVE),
            db_utils.create_share_instance(
                status=constants.STATUS_MANAGING, share_id='fake_id',
                replica_state=constants.REPLICA_STATE_ACTIVE),
        ]

        share1 = db_utils.create_share(instances=instance_list)
        share2 = db_utils.create_share(instances=list(reversed(instance_list)))

        self.assertEqual(
            constants.STATUS_ERROR, share1.instance['status'])
        self.assertEqual(
            constants.STATUS_ERROR, share2.instance['status'])

    def test_access_rules_status_no_instances(self):
        share = db_utils.create_share(instances=[])

        self.assertEqual(constants.STATUS_ACTIVE, share.access_rules_status)

    @ddt.data(constants.STATUS_ACTIVE, constants.SHARE_INSTANCE_RULES_SYNCING,
              constants.SHARE_INSTANCE_RULES_ERROR)
    def test_access_rules_status(self, access_status):
        instances = [
            db_utils.create_share_instance(
                share_id='fake_id', status=constants.STATUS_ERROR,
                access_rules_status=constants.STATUS_ACTIVE),
            db_utils.create_share_instance(
                share_id='fake_id', status=constants.STATUS_AVAILABLE,
                access_rules_status=constants.STATUS_ACTIVE),
            db_utils.create_share_instance(
                share_id='fake_id', status=constants.STATUS_AVAILABLE,
                access_rules_status=access_status),
        ]

        share = db_utils.create_share(instances=instances)

        self.assertEqual(access_status, share.access_rules_status)


@ddt.ddt
class ShareAccessTestCase(test.TestCase):
    """Testing of SQLAlchemy Share Access related model classes."""

    @ddt.data(constants.ACCESS_STATE_QUEUED_TO_APPLY,
              constants.ACCESS_STATE_ACTIVE, constants.ACCESS_STATE_ERROR,
              constants.ACCESS_STATE_APPLYING)
    def test_share_access_mapping_state(self, expected_status):
        ctxt = context.get_admin_context()

        share = db_utils.create_share()
        share_instances = [
            share.instance,
            db_utils.create_share_instance(share_id=share['id']),
            db_utils.create_share_instance(share_id=share['id']),
            db_utils.create_share_instance(share_id=share['id']),
        ]
        access_rule = db_utils.create_access(share_id=share['id'])

        # Update the access mapping states
        db_api.share_instance_access_update(
            ctxt, access_rule['id'], share_instances[0]['id'],
            {'state': constants.ACCESS_STATE_ACTIVE})
        db_api.share_instance_access_update(
            ctxt, access_rule['id'], share_instances[1]['id'],
            {'state': expected_status})
        db_api.share_instance_access_update(
            ctxt, access_rule['id'], share_instances[2]['id'],
            {'state': constants.ACCESS_STATE_ACTIVE})
        db_api.share_instance_access_update(
            ctxt, access_rule['id'], share_instances[3]['id'],
            {'deleted': 'True', 'state': constants.STATUS_DELETED})

        access_rule = db_api.share_access_get(ctxt, access_rule['id'])

        self.assertEqual(expected_status, access_rule['state'])


class ShareSnapshotTestCase(test.TestCase):
    """Testing of SQLAlchemy ShareSnapshot model class."""

    def test_instance_and_proxified_properties(self):

        in_sync_replica_instance = db_utils.create_share_instance(
            status=constants.STATUS_AVAILABLE, share_id='fake_id',
            replica_state=constants.REPLICA_STATE_IN_SYNC)
        active_replica_instance = db_utils.create_share_instance(
            status=constants.STATUS_AVAILABLE, share_id='fake_id',
            replica_state=constants.REPLICA_STATE_ACTIVE)
        out_of_sync_replica_instance = db_utils.create_share_instance(
            status=constants.STATUS_ERROR, share_id='fake_id',
            replica_state=constants.REPLICA_STATE_OUT_OF_SYNC)
        non_replica_instance = db_utils.create_share_instance(
            status=constants.STATUS_CREATING, share_id='fake_id')
        share_instances = [
            in_sync_replica_instance, active_replica_instance,
            out_of_sync_replica_instance, non_replica_instance,
        ]
        share = db_utils.create_share(instances=share_instances)
        snapshot_instance_list = [
            db_utils.create_snapshot_instance(
                'fake_snapshot_id',
                status=constants.STATUS_CREATING,
                share_instance_id=out_of_sync_replica_instance['id']),
            db_utils.create_snapshot_instance(
                'fake_snapshot_id',
                status=constants.STATUS_ERROR,
                share_instance_id=in_sync_replica_instance['id']),
            db_utils.create_snapshot_instance(
                'fake_snapshot_id',
                status=constants.STATUS_AVAILABLE,
                provider_location='hogsmeade:snapshot1',
                progress='87%',
                share_instance_id=active_replica_instance['id']),
            db_utils.create_snapshot_instance(
                'fake_snapshot_id',
                status=constants.STATUS_MANAGING,
                share_instance_id=non_replica_instance['id']),
        ]
        snapshot = db_utils.create_snapshot(
            id='fake_snapshot_id', share_id=share['id'],
            instances=snapshot_instance_list)

        # Proxified properties
        self.assertEqual(constants.STATUS_AVAILABLE, snapshot['status'])
        self.assertEqual(constants.STATUS_ERROR, snapshot['aggregate_status'])
        self.assertEqual('hogsmeade:snapshot1', snapshot['provider_location'])
        self.assertEqual('87%', snapshot['progress'])

        # Snapshot properties
        expected_share_name = '-'.join(['share', share['id']])
        self.assertEqual(expected_share_name, snapshot['share_name'])
        self.assertEqual(active_replica_instance['id'],
                         snapshot['instance']['share_instance_id'])



# Copyright 2013 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

# SHARE AND GENERAL STATUSES
STATUS_CREATING = 'creating'
STATUS_DELETING = 'deleting'
STATUS_DELETED = 'deleted'
STATUS_ERROR = 'error'
STATUS_ERROR_DELETING = 'error_deleting'
STATUS_AVAILABLE = 'available'
STATUS_INACTIVE = 'inactive'
STATUS_MANAGING = 'manage_starting'
STATUS_MANAGE_ERROR = 'manage_error'
STATUS_UNMANAGING = 'unmanage_starting'
STATUS_UNMANAGE_ERROR = 'unmanage_error'
STATUS_UNMANAGED = 'unmanaged'
STATUS_EXTENDING = 'extending'
STATUS_EXTENDING_ERROR = 'extending_error'
STATUS_SHRINKING = 'shrinking'
STATUS_SHRINKING_ERROR = 'shrinking_error'
STATUS_MIGRATING = 'migrating'
STATUS_MIGRATING_TO = 'migrating_to'
STATUS_SHRINKING_POSSIBLE_DATA_LOSS_ERROR = (
    'shrinking_possible_data_loss_error'
)
STATUS_REPLICATION_CHANGE = 'replication_change'
STATUS_RESTORING = 'restoring'
STATUS_REVERTING = 'reverting'
STATUS_REVERTING_ERROR = 'reverting_error'

# Access rule states
ACCESS_STATE_QUEUED_TO_APPLY = 'queued_to_apply'
ACCESS_STATE_QUEUED_TO_DENY = 'queued_to_deny'
ACCESS_STATE_APPLYING = 'applying'
ACCESS_STATE_DENYING = 'denying'
ACCESS_STATE_ACTIVE = 'active'
ACCESS_STATE_ERROR = 'error'
ACCESS_STATE_DELETED = 'deleted'

# Share instance "access_rules_status" field values
SHARE_INSTANCE_RULES_SYNCING = 'syncing'
SHARE_INSTANCE_RULES_ERROR = 'error'

# States/statuses for multiple resources
STATUS_NEW = 'new'
STATUS_OUT_OF_SYNC = 'out_of_sync'
STATUS_ACTIVE = 'active'

ACCESS_RULES_STATES = (
    ACCESS_STATE_QUEUED_TO_APPLY,
    ACCESS_STATE_QUEUED_TO_DENY,
    ACCESS_STATE_APPLYING,
    ACCESS_STATE_DENYING,
    ACCESS_STATE_ACTIVE,
    ACCESS_STATE_ERROR,
    ACCESS_STATE_DELETED,
)

TASK_STATE_MIGRATION_STARTING = 'migration_starting'
TASK_STATE_MIGRATION_IN_PROGRESS = 'migration_in_progress'
TASK_STATE_MIGRATION_COMPLETING = 'migration_completing'
TASK_STATE_MIGRATION_SUCCESS = 'migration_success'
TASK_STATE_MIGRATION_ERROR = 'migration_error'
TASK_STATE_MIGRATION_CANCELLED = 'migration_cancelled'
TASK_STATE_MIGRATION_DRIVER_STARTING = 'migration_driver_starting'
TASK_STATE_MIGRATION_DRIVER_IN_PROGRESS = 'migration_driver_in_progress'
TASK_STATE_MIGRATION_DRIVER_PHASE1_DONE = 'migration_driver_phase1_done'
TASK_STATE_DATA_COPYING_STARTING = 'data_copying_starting'
TASK_STATE_DATA_COPYING_IN_PROGRESS = 'data_copying_in_progress'
TASK_STATE_DATA_COPYING_COMPLETING = 'data_copying_completing'
TASK_STATE_DATA_COPYING_COMPLETED = 'data_copying_completed'
TASK_STATE_DATA_COPYING_CANCELLED = 'data_copying_cancelled'
TASK_STATE_DATA_COPYING_ERROR = 'data_copying_error'

BUSY_TASK_STATES = (
    TASK_STATE_MIGRATION_STARTING,
    TASK_STATE_MIGRATION_IN_PROGRESS,
    TASK_STATE_MIGRATION_COMPLETING,
    TASK_STATE_MIGRATION_DRIVER_STARTING,
    TASK_STATE_MIGRATION_DRIVER_IN_PROGRESS,
    TASK_STATE_MIGRATION_DRIVER_PHASE1_DONE,
    TASK_STATE_DATA_COPYING_STARTING,
    TASK_STATE_DATA_COPYING_IN_PROGRESS,
    TASK_STATE_DATA_COPYING_COMPLETING,
    TASK_STATE_DATA_COPYING_COMPLETED,
)

BUSY_COPYING_STATES = (
    TASK_STATE_DATA_COPYING_STARTING,
    TASK_STATE_DATA_COPYING_IN_PROGRESS,
    TASK_STATE_DATA_COPYING_COMPLETING,
)

TRANSITIONAL_STATUSES = (
    STATUS_CREATING, STATUS_DELETING,
    STATUS_MANAGING, STATUS_UNMANAGING,
    STATUS_EXTENDING, STATUS_SHRINKING,
    STATUS_MIGRATING, STATUS_MIGRATING_TO,
    STATUS_RESTORING, STATUS_REVERTING,
)

INVALID_SHARE_INSTANCE_STATUSES_FOR_ACCESS_RULE_UPDATES = (
    TRANSITIONAL_STATUSES + (STATUS_ERROR,)
)

SUPPORTED_SHARE_PROTOCOLS = (
    'NFS', 'CIFS', 'GLUSTERFS', 'HDFS', 'CEPHFS', 'MAPRFS')

SECURITY_SERVICES_ALLOWED_TYPES = ['active_directory', 'ldap', 'kerberos']

LIKE_FILTER = ['name~', 'description~']

NFS_EXPORTS_FILE = '/etc/exports'
NFS_EXPORTS_FILE_TEMP = '/var/lib/nfs/etab'

MOUNT_FILE = '/etc/fstab'
MOUNT_FILE_TEMP = '/etc/mtab'

# Below represented ports are ranges (from, to)
CIFS_PORTS = (
    ("tcp", (445, 445)),
    ("tcp", (137, 139)),
    ("udp", (137, 139)),
    ("udp", (445, 445)),
)
NFS_PORTS = (
    ("tcp", (2049, 2049)),
    ("udp", (2049, 2049)),
)
SSH_PORTS = (
    ("tcp", (22, 22)),
)
PING_PORTS = (
    ("icmp", (-1, -1)),
)
WINRM_PORTS = (
    ("tcp", (5985, 5986)),
)

SERVICE_INSTANCE_SECGROUP_DATA = (
    CIFS_PORTS + NFS_PORTS + PING_PORTS + WINRM_PORTS)

ACCESS_LEVEL_RW = 'rw'
ACCESS_LEVEL_RO = 'ro'

ACCESS_LEVELS = (
    ACCESS_LEVEL_RW,
    ACCESS_LEVEL_RO,
)

TASK_STATE_STATUSES = (
    TASK_STATE_MIGRATION_STARTING,
    TASK_STATE_MIGRATION_IN_PROGRESS,
    TASK_STATE_MIGRATION_COMPLETING,
    TASK_STATE_MIGRATION_SUCCESS,
    TASK_STATE_MIGRATION_ERROR,
    TASK_STATE_MIGRATION_CANCELLED,
    TASK_STATE_MIGRATION_DRIVER_IN_PROGRESS,
    TASK_STATE_MIGRATION_DRIVER_PHASE1_DONE,
    TASK_STATE_DATA_COPYING_STARTING,
    TASK_STATE_DATA_COPYING_IN_PROGRESS,
    TASK_STATE_DATA_COPYING_COMPLETING,
    TASK_STATE_DATA_COPYING_COMPLETED,
    TASK_STATE_DATA_COPYING_CANCELLED,
    TASK_STATE_DATA_COPYING_ERROR,
    None,
)

REPLICA_STATE_ACTIVE = 'active'
REPLICA_STATE_IN_SYNC = 'in_sync'
REPLICA_STATE_OUT_OF_SYNC = 'out_of_sync'

REPLICATION_TYPE_READABLE = 'readable'
REPLICATION_TYPE_WRITABLE = 'writable'
REPLICATION_TYPE_DR = 'dr'


class ExtraSpecs(object):

    # Extra specs key names
    DRIVER_HANDLES_SHARE_SERVERS = "driver_handles_share_servers"
    SNAPSHOT_SUPPORT = "snapshot_support"
    REPLICATION_TYPE_SPEC = "replication_type"
    CREATE_SHARE_FROM_SNAPSHOT_SUPPORT = "create_share_from_snapshot_support"
    REVERT_TO_SNAPSHOT_SUPPORT = "revert_to_snapshot_support"
    MOUNT_SNAPSHOT_SUPPORT = "mount_snapshot_support"
    AVAILABILITY_ZONES = "availability_zones"

    # Extra specs containers
    REQUIRED = (
        DRIVER_HANDLES_SHARE_SERVERS,
    )

    OPTIONAL = (
        SNAPSHOT_SUPPORT,
        CREATE_SHARE_FROM_SNAPSHOT_SUPPORT,
        REVERT_TO_SNAPSHOT_SUPPORT,
        REPLICATION_TYPE_SPEC,
        MOUNT_SNAPSHOT_SUPPORT,
        AVAILABILITY_ZONES,
    )

    # NOTE(cknight): Some extra specs are necessary parts of the Manila API and
    # should be visible to non-admin users. REQUIRED specs are user-visible, as
    # are a handful of community-agreed standardized OPTIONAL ones.
    TENANT_VISIBLE = REQUIRED + OPTIONAL

    BOOLEAN = (
        DRIVER_HANDLES_SHARE_SERVERS,
        SNAPSHOT_SUPPORT,
        CREATE_SHARE_FROM_SNAPSHOT_SUPPORT,
        REVERT_TO_SNAPSHOT_SUPPORT,
        MOUNT_SNAPSHOT_SUPPORT,
    )

    # NOTE(cknight): Some extra specs are optional, but a nominal (typically
    # False, but may be non-boolean) default value for each is still needed
    # when creating shares.
    INFERRED_OPTIONAL_MAP = {
        SNAPSHOT_SUPPORT: False,
        CREATE_SHARE_FROM_SNAPSHOT_SUPPORT: False,
        REVERT_TO_SNAPSHOT_SUPPORT: False,
        MOUNT_SNAPSHOT_SUPPORT: False,
    }

    REPLICATION_TYPES = ('writable', 'readable', 'dr')
